@@ -23,16 +23,31 @@
 #include <config.h>
 #include <inttypes.h>
 #include "char-device.h"
-#include "red-channel.h"
+#include "red-client.h"
 #include "reds.h"
+#include "glib-compat.h"
 
 #define CHAR_DEVICE_WRITE_TO_TIMEOUT 100
 #define RED_CHAR_DEVICE_WAIT_TOKENS_TIMEOUT 30000
 #define MAX_POOL_SIZE (10 * 64 * 1024)
 
+typedef enum {
+    WRITE_BUFFER_ORIGIN_NONE,
+    WRITE_BUFFER_ORIGIN_CLIENT,
+    WRITE_BUFFER_ORIGIN_SERVER,
+    WRITE_BUFFER_ORIGIN_SERVER_NO_TOKEN,
+} WriteBufferOrigin;
+
+struct RedCharDeviceWriteBufferPrivate {
+    RedClient *client; /* The client that sent the message to the device.
+                          NULL if the server created the message */
+    WriteBufferOrigin origin;
+    uint32_t token_price;
+    uint32_t refs;
+};
+
 typedef struct RedCharDeviceClient RedCharDeviceClient;
 struct RedCharDeviceClient {
-    RingItem link;
     RedCharDevice *dev;
     RedClient *client;
     int do_flow_control;
@@ -50,16 +65,15 @@ struct RedCharDevicePrivate {
     int active; /* has read/write been performed since the device was started */
     int wait_for_migrate_data;
 
-    Ring write_queue;
-    Ring write_bufs_pool;
+    GQueue write_queue;
+    GQueue write_bufs_pool;
     uint64_t cur_pool_size;
     RedCharDeviceWriteBuffer *cur_write_buf;
     uint8_t *cur_write_buf_pos;
     SpiceTimer *write_to_dev_timer;
     uint64_t num_self_tokens;
 
-    Ring clients; /* list of RedCharDeviceClient */
-    uint32_t num_clients;
+    GList *clients; /* list of RedCharDeviceClient */
 
     uint64_t client_tokens_interval; /* frequency of returning tokens to the client */
     SpiceCharDeviceInstance *sin;
@@ -67,18 +81,9 @@ struct RedCharDevicePrivate {
     int during_read_from_device;
     int during_write_to_device;
 
-    void *opaque;
     SpiceServer *reds;
 };
 
-enum {
-    WRITE_BUFFER_ORIGIN_NONE,
-    WRITE_BUFFER_ORIGIN_CLIENT,
-    WRITE_BUFFER_ORIGIN_SERVER,
-    WRITE_BUFFER_ORIGIN_SERVER_NO_TOKEN,
-};
-
-
 G_DEFINE_TYPE(RedCharDevice, red_char_device, G_TYPE_OBJECT)
 
 #define RED_CHAR_DEVICE_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), RED_TYPE_CHAR_DEVICE, RedCharDevicePrivate))
@@ -89,7 +94,6 @@ enum {
     PROP_SPICE_SERVER,
     PROP_CLIENT_TOKENS_INTERVAL,
     PROP_SELF_TOKENS,
-    PROP_OPAQUE
 };
 
 static void red_char_device_write_buffer_unref(RedCharDeviceWriteBuffer *write_buf);
@@ -100,7 +104,7 @@ red_char_device_read_one_msg_from_device(RedCharDevice *dev)
 {
    RedCharDeviceClass *klass = RED_CHAR_DEVICE_GET_CLASS(dev);
 
-   return klass->read_one_msg_from_device(dev->priv->sin, dev->priv->opaque);
+   return klass->read_one_msg_from_device(dev, dev->priv->sin);
 }
 
 static void
@@ -110,7 +114,7 @@ red_char_device_send_msg_to_client(RedCharDevice *dev,
 {
    RedCharDeviceClass *klass = RED_CHAR_DEVICE_GET_CLASS(dev);
 
-   klass->send_msg_to_client(msg, client, dev->priv->opaque);
+   klass->send_msg_to_client(dev, msg, client);
 }
 
 static void
@@ -120,7 +124,7 @@ red_char_device_send_tokens_to_client(RedCharDevice *dev,
 {
    RedCharDeviceClass *klass = RED_CHAR_DEVICE_GET_CLASS(dev);
 
-   klass->send_tokens_to_client(client, tokens, dev->priv->opaque);
+   klass->send_tokens_to_client(dev, client, tokens);
 }
 
 static void
@@ -129,7 +133,7 @@ red_char_device_on_free_self_token(RedCharDevice *dev)
    RedCharDeviceClass *klass = RED_CHAR_DEVICE_GET_CLASS(dev);
 
    if (klass->on_free_self_token != NULL) {
-       klass->on_free_self_token(dev->priv->opaque);
+       klass->on_free_self_token(dev);
    }
 }
 
@@ -138,7 +142,7 @@ red_char_device_remove_client(RedCharDevice *dev, RedClient *client)
 {
    RedCharDeviceClass *klass = RED_CHAR_DEVICE_GET_CLASS(dev);
 
-   klass->remove_client(client, dev->priv->opaque);
+   klass->remove_client(dev, client);
 }
 
 static void red_char_device_write_buffer_free(RedCharDeviceWriteBuffer *buf)
@@ -150,28 +154,23 @@ static void red_char_device_write_buffer_free(RedCharDeviceWriteBuffer *buf)
     free(buf);
 }
 
-static void write_buffers_queue_free(Ring *write_queue)
+static void write_buffers_queue_free(GQueue *write_queue)
 {
-    while (!ring_is_empty(write_queue)) {
-        RingItem *item = ring_get_tail(write_queue);
-        RedCharDeviceWriteBuffer *buf;
-
-        ring_remove(item);
-        buf = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
+    RedCharDeviceWriteBuffer *buf;
+    while ((buf = g_queue_pop_tail(write_queue)))
         red_char_device_write_buffer_free(buf);
-    }
 }
 
 static void red_char_device_write_buffer_pool_add(RedCharDevice *dev,
                                                   RedCharDeviceWriteBuffer *buf)
 {
-    if (buf->refs == 1 &&
+    if (buf->priv->refs == 1 &&
         dev->priv->cur_pool_size < MAX_POOL_SIZE) {
         buf->buf_used = 0;
-        buf->origin = WRITE_BUFFER_ORIGIN_NONE;
-        buf->client = NULL;
+        buf->priv->origin = WRITE_BUFFER_ORIGIN_NONE;
+        buf->priv->client = NULL;
         dev->priv->cur_pool_size += buf->buf_size;
-        ring_add(&dev->priv->write_bufs_pool, &buf->link);
+        g_queue_push_head(&dev->priv->write_bufs_pool, buf);
         return;
     }
 
@@ -182,36 +181,35 @@ static void red_char_device_write_buffer_pool_add(RedCharDevice *dev,
 static void red_char_device_client_free(RedCharDevice *dev,
                                         RedCharDeviceClient *dev_client)
 {
-    RingItem *item, *next;
+    GList *l, *next;
 
-    if (dev_client->wait_for_tokens_timer) {
-        reds_core_timer_remove(dev->priv->reds, dev_client->wait_for_tokens_timer);
-        dev_client->wait_for_tokens_timer = NULL;
-    }
+    reds_core_timer_remove(dev->priv->reds, dev_client->wait_for_tokens_timer);
+    dev_client->wait_for_tokens_timer = NULL;
 
     g_queue_free_full(dev_client->send_queue, (GDestroyNotify)red_pipe_item_unref);
 
     /* remove write buffers that are associated with the client */
-    spice_debug("write_queue_is_empty %d", ring_is_empty(&dev->priv->write_queue) && !dev->priv->cur_write_buf);
-    RING_FOREACH_SAFE(item, next, &dev->priv->write_queue) {
-        RedCharDeviceWriteBuffer *write_buf;
+    spice_debug("write_queue_is_empty %d", g_queue_is_empty(&dev->priv->write_queue) && !dev->priv->cur_write_buf);
+    l = g_queue_peek_head_link(&dev->priv->write_queue);
+    while (l) {
+        RedCharDeviceWriteBuffer *write_buf = l->data;
+        next = l->next;
 
-        write_buf = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
-        if (write_buf->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
-            write_buf->client == dev_client->client) {
-            ring_remove(item);
+        if (write_buf->priv->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
+            write_buf->priv->client == dev_client->client) {
+            g_queue_delete_link(&dev->priv->write_queue, l);
             red_char_device_write_buffer_pool_add(dev, write_buf);
         }
+        l = next;
     }
 
-    if (dev->priv->cur_write_buf && dev->priv->cur_write_buf->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
-        dev->priv->cur_write_buf->client == dev_client->client) {
-        dev->priv->cur_write_buf->origin = WRITE_BUFFER_ORIGIN_NONE;
-        dev->priv->cur_write_buf->client = NULL;
+    if (dev->priv->cur_write_buf && dev->priv->cur_write_buf->priv->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
+        dev->priv->cur_write_buf->priv->client == dev_client->client) {
+        dev->priv->cur_write_buf->priv->origin = WRITE_BUFFER_ORIGIN_NONE;
+        dev->priv->cur_write_buf->priv->client = NULL;
     }
 
-    dev->priv->num_clients--;
-    ring_remove(&dev_client->link);
+    dev->priv->clients = g_list_remove(dev->priv->clients, dev_client);
     free(dev_client);
 }
 
@@ -225,12 +223,11 @@ static void red_char_device_handle_client_overflow(RedCharDeviceClient *dev_clie
 static RedCharDeviceClient *red_char_device_client_find(RedCharDevice *dev,
                                                         RedClient *client)
 {
-    RingItem *item;
+    GList *item;
 
-    RING_FOREACH(item, &dev->priv->clients) {
-        RedCharDeviceClient *dev_client;
+    for (item = dev->priv->clients; item != NULL; item = item->next) {
+        RedCharDeviceClient *dev_client = item->data;
 
-        dev_client = SPICE_CONTAINEROF(item, RedCharDeviceClient, link);
         if (dev_client->client == client) {
             return dev_client;
         }
@@ -256,13 +253,11 @@ static int red_char_device_can_send_to_client(RedCharDeviceClient *dev_client)
 
 static uint64_t red_char_device_max_send_tokens(RedCharDevice *dev)
 {
-    RingItem *item;
+    GList *item;
     uint64_t max = 0;
 
-    RING_FOREACH(item, &dev->priv->clients) {
-        RedCharDeviceClient *dev_client;
-
-        dev_client = SPICE_CONTAINEROF(item, RedCharDeviceClient, link);
+    for (item = dev->priv->clients; item != NULL; item = item->next) {
+        RedCharDeviceClient *dev_client = item->data;
 
         if (!dev_client->do_flow_control) {
             max = ~0;
@@ -298,12 +293,13 @@ static void red_char_device_add_msg_to_client_queue(RedCharDeviceClient *dev_cli
 static void red_char_device_send_msg_to_clients(RedCharDevice *dev,
                                                 RedPipeItem *msg)
 {
-    RingItem *item, *next;
+    GList *l;
 
-    RING_FOREACH_SAFE(item, next, &dev->priv->clients) {
-        RedCharDeviceClient *dev_client;
+    l = dev->priv->clients;
+    while (l) {
+        GList *next = l->next;
+        RedCharDeviceClient *dev_client = l->data;
 
-        dev_client = SPICE_CONTAINEROF(item, RedCharDeviceClient, link);
         if (red_char_device_can_send_to_client(dev_client)) {
             dev_client->num_send_tokens--;
             spice_assert(g_queue_is_empty(dev_client->send_queue));
@@ -313,10 +309,11 @@ static void red_char_device_send_msg_to_clients(RedCharDevice *dev,
         } else {
             red_char_device_add_msg_to_client_queue(dev_client, msg);
         }
+        l = next;
     }
 }
 
-static int red_char_device_read_from_device(RedCharDevice *dev)
+static bool red_char_device_read_from_device(RedCharDevice *dev)
 {
     uint64_t max_send_tokens;
     int did_read = FALSE;
@@ -341,7 +338,7 @@ static int red_char_device_read_from_device(RedCharDevice *dev)
      * Reading from the device only in case at least one of the clients have a free token.
      * All messages will be discarded if no client is attached to the device
      */
-    while ((max_send_tokens || ring_is_empty(&dev->priv->clients)) && dev->priv->running) {
+    while ((max_send_tokens || (dev->priv->clients == NULL)) && dev->priv->running) {
         RedPipeItem *msg;
 
         msg = red_char_device_read_one_msg_from_device(dev);
@@ -483,13 +480,10 @@ static int red_char_device_write_to_device(RedCharDevice *dev)
         uint32_t write_len;
 
         if (!dev->priv->cur_write_buf) {
-            RingItem *item = ring_get_tail(&dev->priv->write_queue);
-            if (!item) {
+            dev->priv->cur_write_buf = g_queue_pop_tail(&dev->priv->write_queue);
+            if (!dev->priv->cur_write_buf)
                 break;
-            }
-            dev->priv->cur_write_buf = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
             dev->priv->cur_write_buf_pos = dev->priv->cur_write_buf->buf;
-            ring_remove(item);
         }
 
         write_len = dev->priv->cur_write_buf->buf + dev->priv->cur_write_buf->buf_used -
@@ -519,7 +513,7 @@ static int red_char_device_write_to_device(RedCharDevice *dev)
                                       CHAR_DEVICE_WRITE_TO_TIMEOUT);
             }
         } else {
-            spice_assert(ring_is_empty(&dev->priv->write_queue));
+            spice_assert(g_queue_is_empty(&dev->priv->write_queue));
         }
         dev->priv->active = dev->priv->active || total;
     }
@@ -540,21 +534,25 @@ static void red_char_device_write_retry(void *opaque)
 
 static RedCharDeviceWriteBuffer *__red_char_device_write_buffer_get(
     RedCharDevice *dev, RedClient *client,
-    int size, int origin, int migrated_data_tokens)
+    int size, WriteBufferOrigin origin, int migrated_data_tokens)
 {
-    RingItem *item;
     RedCharDeviceWriteBuffer *ret;
 
     if (origin == WRITE_BUFFER_ORIGIN_SERVER && !dev->priv->num_self_tokens) {
         return NULL;
     }
 
-    if ((item = ring_get_tail(&dev->priv->write_bufs_pool))) {
-        ret = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
-        ring_remove(item);
+    ret = g_queue_pop_tail(&dev->priv->write_bufs_pool);
+    if (ret) {
         dev->priv->cur_pool_size -= ret->buf_size;
     } else {
-        ret = spice_new0(RedCharDeviceWriteBuffer, 1);
+        struct RedCharDeviceWriteBufferFull {
+            RedCharDeviceWriteBuffer buffer;
+            RedCharDeviceWriteBufferPrivate priv;
+        } *write_buf;
+        write_buf = spice_new0(struct RedCharDeviceWriteBufferFull, 1);
+        ret = &write_buf->buffer;
+        ret->priv = &write_buf->priv;
     }
 
     spice_assert(!ret->buf_used);
@@ -563,7 +561,7 @@ static RedCharDeviceWriteBuffer *__red_char_device_write_buffer_get(
         ret->buf = spice_realloc(ret->buf, size);
         ret->buf_size = size;
     }
-    ret->origin = origin;
+    ret->priv->origin = origin;
 
     if (origin == WRITE_BUFFER_ORIGIN_CLIENT) {
        spice_assert(client);
@@ -575,7 +573,7 @@ static RedCharDeviceWriteBuffer *__red_char_device_write_buffer_get(
                 red_char_device_handle_client_overflow(dev_client);
                 goto error;
             }
-            ret->client = client;
+            ret->priv->client = client;
             if (!migrated_data_tokens && dev_client->do_flow_control) {
                 dev_client->num_client_tokens--;
             }
@@ -589,12 +587,12 @@ static RedCharDeviceWriteBuffer *__red_char_device_write_buffer_get(
         dev->priv->num_self_tokens--;
     }
 
-    ret->token_price = migrated_data_tokens ? migrated_data_tokens : 1;
-    ret->refs = 1;
+    ret->priv->token_price = migrated_data_tokens ? migrated_data_tokens : 1;
+    ret->priv->refs = 1;
     return ret;
 error:
     dev->priv->cur_pool_size += ret->buf_size;
-    ring_add(&dev->priv->write_bufs_pool, &ret->link);
+    g_queue_push_head(&dev->priv->write_bufs_pool, ret);
     return NULL;
 }
 
@@ -618,7 +616,7 @@ static RedCharDeviceWriteBuffer *red_char_device_write_buffer_ref(RedCharDeviceW
 {
     spice_assert(write_buf);
 
-    write_buf->refs++;
+    write_buf->priv->refs++;
     return write_buf;
 }
 
@@ -626,8 +624,8 @@ static void red_char_device_write_buffer_unref(RedCharDeviceWriteBuffer *write_b
 {
     spice_assert(write_buf);
 
-    write_buf->refs--;
-    if (write_buf->refs == 0)
+    write_buf->priv->refs--;
+    if (write_buf->priv->refs == 0)
         red_char_device_write_buffer_free(write_buf);
 }
 
@@ -636,14 +634,14 @@ void red_char_device_write_buffer_add(RedCharDevice *dev,
 {
     spice_assert(dev);
     /* caller shouldn't add buffers for client that was removed */
-    if (write_buf->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
-        !red_char_device_client_find(dev, write_buf->client)) {
-        spice_printerr("client not found: dev %p client %p", dev, write_buf->client);
+    if (write_buf->priv->origin == WRITE_BUFFER_ORIGIN_CLIENT &&
+        !red_char_device_client_find(dev, write_buf->priv->client)) {
+        spice_printerr("client not found: dev %p client %p", dev, write_buf->priv->client);
         red_char_device_write_buffer_pool_add(dev, write_buf);
         return;
     }
 
-    ring_add(&dev->priv->write_queue, &write_buf->link);
+    g_queue_push_head(&dev->priv->write_queue, write_buf);
     red_char_device_write_to_device(dev);
 }
 
@@ -656,11 +654,10 @@ void red_char_device_write_buffer_release(RedCharDevice *dev,
     }
     *p_write_buf = NULL;
 
-    int buf_origin = write_buf->origin;
-    uint32_t buf_token_price = write_buf->token_price;
-    RedClient *client = write_buf->client;
+    WriteBufferOrigin buf_origin = write_buf->priv->origin;
+    uint32_t buf_token_price = write_buf->priv->token_price;
+    RedClient *client = write_buf->priv->client;
 
-    spice_assert(!ring_item_is_linked(&write_buf->link));
     if (!dev) {
         spice_printerr("no device. write buffer is freed");
         red_char_device_write_buffer_free(write_buf);
@@ -698,11 +695,6 @@ void red_char_device_reset_dev_instance(RedCharDevice *dev,
     g_object_notify(G_OBJECT(dev), "sin");
 }
 
-void *red_char_device_opaque_get(RedCharDevice *dev)
-{
-    return dev->priv->opaque;
-}
-
 void red_char_device_destroy(RedCharDevice *char_dev)
 {
     g_return_if_fail(RED_IS_CHAR_DEVICE(char_dev));
@@ -723,8 +715,10 @@ static RedCharDeviceClient *red_char_device_client_new(RedClient *client,
     dev_client->max_send_queue_size = max_send_queue_size;
     dev_client->do_flow_control = do_flow_control;
     if (do_flow_control) {
+        RedsState *reds = red_client_get_server(client);
+
         dev_client->wait_for_tokens_timer =
-            reds_core_timer_add(client->reds, device_client_wait_for_tokens_timeout,
+            reds_core_timer_add(reds, device_client_wait_for_tokens_timeout,
                                 dev_client);
         if (!dev_client->wait_for_tokens_timer) {
             spice_error("failed to create wait for tokens timer");
@@ -739,20 +733,20 @@ static RedCharDeviceClient *red_char_device_client_new(RedClient *client,
     return dev_client;
 }
 
-int red_char_device_client_add(RedCharDevice *dev,
-                               RedClient *client,
-                               int do_flow_control,
-                               uint32_t max_send_queue_size,
-                               uint32_t num_client_tokens,
-                               uint32_t num_send_tokens,
-                               int wait_for_migrate_data)
+bool red_char_device_client_add(RedCharDevice *dev,
+                                RedClient *client,
+                                int do_flow_control,
+                                uint32_t max_send_queue_size,
+                                uint32_t num_client_tokens,
+                                uint32_t num_send_tokens,
+                                int wait_for_migrate_data)
 {
     RedCharDeviceClient *dev_client;
 
     spice_assert(dev);
     spice_assert(client);
 
-    if (wait_for_migrate_data && (dev->priv->num_clients > 0 || dev->priv->active)) {
+    if (wait_for_migrate_data && (dev->priv->clients != NULL || dev->priv->active)) {
         spice_warning("can't restore device %p from migration data. The device "
                       "has already been active", dev);
         return FALSE;
@@ -766,8 +760,7 @@ int red_char_device_client_add(RedCharDevice *dev,
                                             num_client_tokens,
                                             num_send_tokens);
     dev_client->dev = dev;
-    ring_add(&dev->priv->clients, &dev_client->link);
-    dev->priv->num_clients++;
+    dev->priv->clients = g_list_prepend(dev->priv->clients, dev_client);
     /* Now that we have a client, forward any pending device data */
     red_char_device_wakeup(dev);
     return TRUE;
@@ -787,12 +780,12 @@ void red_char_device_client_remove(RedCharDevice *dev,
     }
     red_char_device_client_free(dev, dev_client);
     if (dev->priv->wait_for_migrate_data) {
-        spice_assert(dev->priv->num_clients == 0);
+        spice_assert(dev->priv->clients == NULL);
         dev->priv->wait_for_migrate_data  = FALSE;
         red_char_device_read_from_device(dev);
     }
 
-    if (dev->priv->num_clients == 0) {
+    if (dev->priv->clients == NULL) {
         spice_debug("client removed, memory pool will be freed (%"PRIu64" bytes)", dev->priv->cur_pool_size);
         write_buffers_queue_free(&dev->priv->write_bufs_pool);
         dev->priv->cur_pool_size = 0;
@@ -827,26 +820,20 @@ void red_char_device_stop(RedCharDevice *dev)
 
 void red_char_device_reset(RedCharDevice *dev)
 {
-    RingItem *client_item;
+    GList *client_item;
+    RedCharDeviceWriteBuffer *buf;
 
     red_char_device_stop(dev);
     dev->priv->wait_for_migrate_data = FALSE;
     spice_debug("char device %p", dev);
-    while (!ring_is_empty(&dev->priv->write_queue)) {
-        RingItem *item = ring_get_tail(&dev->priv->write_queue);
-        RedCharDeviceWriteBuffer *buf;
-
-        ring_remove(item);
-        buf = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
-        /* tracking the tokens */
+    while ((buf = g_queue_pop_tail(&dev->priv->write_queue))) {
         red_char_device_write_buffer_release(dev, &buf);
     }
     red_char_device_write_buffer_release(dev, &dev->priv->cur_write_buf);
 
-    RING_FOREACH(client_item, &dev->priv->clients) {
-        RedCharDeviceClient *dev_client;
+    for (client_item = dev->priv->clients; client_item != NULL; client_item = client_item->next) {
+        RedCharDeviceClient *dev_client = client_item->data;
 
-        dev_client = SPICE_CONTAINEROF(client_item, RedCharDeviceClient, link);
         spice_debug("send_queue_empty %d", g_queue_is_empty(dev_client->send_queue));
         dev_client->num_send_tokens += g_queue_get_length(dev_client->send_queue);
         g_queue_foreach(dev_client->send_queue, (GFunc)red_pipe_item_unref, NULL);
@@ -875,7 +862,7 @@ void red_char_device_migrate_data_marshall_empty(SpiceMarshaller *m)
 {
     SpiceMigrateDataCharDevice *mig_data;
 
-    spice_debug(NULL);
+    spice_debug("trace");
     mig_data = (SpiceMigrateDataCharDevice *)spice_marshaller_reserve_space(m,
                                                                             sizeof(*mig_data));
     memset(mig_data, 0, sizeof(*mig_data));
@@ -894,16 +881,14 @@ void red_char_device_migrate_data_marshall(RedCharDevice *dev,
                                            SpiceMarshaller *m)
 {
     RedCharDeviceClient *dev_client;
-    RingItem *item;
+    GList *item;
     uint32_t *write_to_dev_size_ptr;
     uint32_t *write_to_dev_tokens_ptr;
     SpiceMarshaller *m2;
 
     /* multi-clients are not supported */
-    spice_assert(dev->priv->num_clients == 1);
-    dev_client = SPICE_CONTAINEROF(ring_get_tail(&dev->priv->clients),
-                                   RedCharDeviceClient,
-                                   link);
+    spice_assert(g_list_length(dev->priv->clients) == 1);
+    dev_client = g_list_last(dev->priv->clients)->data;
     /* FIXME: if there were more than one client before the marshalling,
      * it is possible that the send_queue length > 0, and the send data
      * should be migrated as well */
@@ -921,52 +906,50 @@ void red_char_device_migrate_data_marshall(RedCharDevice *dev,
     if (dev->priv->cur_write_buf) {
         uint32_t buf_remaining = dev->priv->cur_write_buf->buf + dev->priv->cur_write_buf->buf_used -
                                  dev->priv->cur_write_buf_pos;
-        spice_marshaller_add_ref_full(m2, dev->priv->cur_write_buf_pos, buf_remaining,
-                                      migrate_data_marshaller_write_buffer_free,
-                                      red_char_device_write_buffer_ref(dev->priv->cur_write_buf)
-                                      );
+        spice_marshaller_add_by_ref_full(m2, dev->priv->cur_write_buf_pos, buf_remaining,
+                                         migrate_data_marshaller_write_buffer_free,
+                                         red_char_device_write_buffer_ref(dev->priv->cur_write_buf)
+                                         );
         *write_to_dev_size_ptr += buf_remaining;
-        if (dev->priv->cur_write_buf->origin == WRITE_BUFFER_ORIGIN_CLIENT) {
-            spice_assert(dev->priv->cur_write_buf->client == dev_client->client);
-            (*write_to_dev_tokens_ptr) += dev->priv->cur_write_buf->token_price;
+        if (dev->priv->cur_write_buf->priv->origin == WRITE_BUFFER_ORIGIN_CLIENT) {
+            spice_assert(dev->priv->cur_write_buf->priv->client == dev_client->client);
+            (*write_to_dev_tokens_ptr) += dev->priv->cur_write_buf->priv->token_price;
         }
     }
 
-    RING_FOREACH_REVERSED(item, &dev->priv->write_queue) {
-        RedCharDeviceWriteBuffer *write_buf;
+    for (item = g_queue_peek_tail_link(&dev->priv->write_queue); item != NULL; item = item->prev) {
+        RedCharDeviceWriteBuffer *write_buf = item->data;
 
-        write_buf = SPICE_CONTAINEROF(item, RedCharDeviceWriteBuffer, link);
-        spice_marshaller_add_ref_full(m2, write_buf->buf, write_buf->buf_used,
-                                      migrate_data_marshaller_write_buffer_free,
-                                      red_char_device_write_buffer_ref(write_buf)
-                                      );
+        spice_marshaller_add_by_ref_full(m2, write_buf->buf, write_buf->buf_used,
+                                         migrate_data_marshaller_write_buffer_free,
+                                         red_char_device_write_buffer_ref(write_buf)
+                                         );
         *write_to_dev_size_ptr += write_buf->buf_used;
-        if (write_buf->origin == WRITE_BUFFER_ORIGIN_CLIENT) {
-            spice_assert(write_buf->client == dev_client->client);
-            (*write_to_dev_tokens_ptr) += write_buf->token_price;
+        if (write_buf->priv->origin == WRITE_BUFFER_ORIGIN_CLIENT) {
+            spice_assert(write_buf->priv->client == dev_client->client);
+            (*write_to_dev_tokens_ptr) += write_buf->priv->token_price;
         }
     }
     spice_debug("migration data dev %p: write_queue size %u tokens %u",
                 dev, *write_to_dev_size_ptr, *write_to_dev_tokens_ptr);
 }
 
-int red_char_device_restore(RedCharDevice *dev,
-                            SpiceMigrateDataCharDevice *mig_data)
+bool red_char_device_restore(RedCharDevice *dev,
+                             SpiceMigrateDataCharDevice *mig_data)
 {
     RedCharDeviceClient *dev_client;
     uint32_t client_tokens_window;
 
-    spice_assert(dev->priv->num_clients == 1 && dev->priv->wait_for_migrate_data);
+    spice_assert(g_list_length(dev->priv->clients) == 1 &&
+                 dev->priv->wait_for_migrate_data);
 
-    dev_client = SPICE_CONTAINEROF(ring_get_tail(&dev->priv->clients),
-                                     RedCharDeviceClient,
-                                     link);
+    dev_client = g_list_last(dev->priv->clients)->data;
     if (mig_data->version > SPICE_MIGRATE_DATA_CHAR_DEVICE_VERSION) {
         spice_error("dev %p error: migration data version %u is bigger than self %u",
                     dev, mig_data->version, SPICE_MIGRATE_DATA_CHAR_DEVICE_VERSION);
         return FALSE;
     }
-    spice_assert(!dev->priv->cur_write_buf && ring_is_empty(&dev->priv->write_queue));
+    spice_assert(!dev->priv->cur_write_buf && g_queue_is_empty(&dev->priv->write_queue));
     spice_assert(mig_data->connected);
 
     client_tokens_window = dev_client->num_client_tokens; /* initial state of tokens */
@@ -1018,10 +1001,9 @@ static void red_char_device_init_device_instance(RedCharDevice *self)
 
     g_return_if_fail(self->priv->reds);
 
-    if (self->priv->write_to_dev_timer) {
-        reds_core_timer_remove(self->priv->reds, self->priv->write_to_dev_timer);
-        self->priv->write_to_dev_timer = NULL;
-    }
+    reds_core_timer_remove(self->priv->reds, self->priv->write_to_dev_timer);
+    self->priv->write_to_dev_timer = NULL;
+
     if (self->priv->sin == NULL) {
        return;
     }
@@ -1062,9 +1044,6 @@ red_char_device_get_property(GObject    *object,
         case PROP_SELF_TOKENS:
             g_value_set_uint64(value, self->priv->num_self_tokens);
             break;
-        case PROP_OPAQUE:
-            g_value_set_pointer(value, self->priv->opaque);
-            break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
     }
@@ -1092,9 +1071,6 @@ red_char_device_set_property(GObject      *object,
         case PROP_SELF_TOKENS:
             self->priv->num_self_tokens = g_value_get_uint64(value);
             break;
-        case PROP_OPAQUE:
-            self->priv->opaque = g_value_get_pointer(value);
-            break;
         default:
             G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
     }
@@ -1115,21 +1091,17 @@ red_char_device_finalize(GObject *object)
 {
     RedCharDevice *self = RED_CHAR_DEVICE(object);
 
-    if (self->priv->write_to_dev_timer) {
-        reds_core_timer_remove(self->priv->reds, self->priv->write_to_dev_timer);
-        self->priv->write_to_dev_timer = NULL;
-    }
+    reds_core_timer_remove(self->priv->reds, self->priv->write_to_dev_timer);
+    self->priv->write_to_dev_timer = NULL;
+
     write_buffers_queue_free(&self->priv->write_queue);
     write_buffers_queue_free(&self->priv->write_bufs_pool);
     self->priv->cur_pool_size = 0;
     red_char_device_write_buffer_free(self->priv->cur_write_buf);
     self->priv->cur_write_buf = NULL;
 
-    while (!ring_is_empty(&self->priv->clients)) {
-        RingItem *item = ring_get_tail(&self->priv->clients);
-        RedCharDeviceClient *dev_client;
-
-        dev_client = SPICE_CONTAINEROF(item, RedCharDeviceClient, link);
+    while (self->priv->clients != NULL) {
+        RedCharDeviceClient *dev_client = self->priv->clients->data;
         red_char_device_client_free(self, dev_client);
     }
     self->priv->running = FALSE;
@@ -1180,14 +1152,6 @@ red_char_device_class_init(RedCharDeviceClass *klass)
                                                         0, G_MAXUINT64, 0,
                                                         G_PARAM_STATIC_STRINGS |
                                                         G_PARAM_READWRITE));
-    g_object_class_install_property(object_class,
-                                    PROP_OPAQUE,
-                                    g_param_spec_pointer("opaque",
-                                                         "opaque",
-                                                         "User data to pass to callbacks",
-                                                      G_PARAM_STATIC_STRINGS |
-                                                      G_PARAM_READWRITE));
-
 }
 
 static void
@@ -1195,9 +1159,8 @@ red_char_device_init(RedCharDevice *self)
 {
     self->priv = RED_CHAR_DEVICE_PRIVATE(self);
 
-    ring_init(&self->priv->write_queue);
-    ring_init(&self->priv->write_bufs_pool);
-    ring_init(&self->priv->clients);
+    g_queue_init(&self->priv->write_queue);
+    g_queue_init(&self->priv->write_bufs_pool);
 
     g_signal_connect(self, "notify::sin", G_CALLBACK(red_char_device_on_sin_changed), NULL);
 }
