@@ -34,14 +34,12 @@ typedef struct CursorItem {
     RedCursorCmd *red_cursor;
 } CursorItem;
 
-G_STATIC_ASSERT(sizeof(CursorItem) <= QXL_CURSUR_DEVICE_DATA_SIZE);
-
 struct CursorChannel
 {
     CommonGraphicsChannel parent;
 
     CursorItem *item;
-    int cursor_visible;
+    bool cursor_visible;
     SpicePoint16 cursor_position;
     uint16_t cursor_trail_length;
     uint16_t cursor_trail_frequency;
@@ -104,7 +102,7 @@ static void cursor_item_unref(CursorItem *item)
 
 }
 
-static void cursor_set_item(CursorChannel *cursor, CursorItem *item)
+static void cursor_channel_set_item(CursorChannel *cursor, CursorItem *item)
 {
     if (cursor->item)
         cursor_item_unref(cursor->item);
@@ -123,23 +121,16 @@ static RedPipeItem *new_cursor_pipe_item(RedChannelClient *rcc, void *data, int
     return &item->base;
 }
 
-typedef struct {
-    void *data;
-    uint32_t size;
-} AddBufInfo;
-
-static void add_buf_from_info(SpiceMarshaller *m, AddBufInfo *info)
+static void marshaller_unref_cursor_item(uint8_t *data, void *opaque)
 {
-    if (info->data) {
-        spice_marshaller_add_ref(m, info->data, info->size);
-    }
+    CursorItem *item = opaque;
+    cursor_item_unref(item);
 }
 
-static void cursor_fill(CursorChannelClient *ccc, SpiceCursor *red_cursor,
-                        CursorItem *cursor, AddBufInfo *addbuf)
+static void cursor_fill(CursorChannelClient *ccc, CursorItem *cursor,
+                        SpiceCursor *red_cursor, SpiceMarshaller *m)
 {
     RedCursorCmd *cursor_cmd;
-    addbuf->data = NULL;
 
     if (!cursor) {
         red_cursor->flags = SPICE_CURSOR_FLAGS_NONE;
@@ -160,8 +151,10 @@ static void cursor_fill(CursorChannelClient *ccc, SpiceCursor *red_cursor,
     }
 
     if (red_cursor->data_size) {
-        addbuf->data = red_cursor->data;
-        addbuf->size = red_cursor->data_size;
+        SpiceMarshaller *m2 = spice_marshaller_get_submarshaller(m);
+        cursor_item_ref(cursor);
+        spice_marshaller_add_by_ref_full(m2, red_cursor->data, red_cursor->data_size,
+                                         marshaller_unref_cursor_item, cursor);
     }
 }
 
@@ -192,20 +185,18 @@ static void red_marshall_cursor_init(CursorChannelClient *ccc, SpiceMarshaller *
     CursorChannel *cursor_channel;
     RedChannelClient *rcc = RED_CHANNEL_CLIENT(ccc);
     SpiceMsgCursorInit msg;
-    AddBufInfo info;
 
     spice_assert(rcc);
     cursor_channel = CURSOR_CHANNEL(red_channel_client_get_channel(rcc));
 
-    red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INIT, NULL);
+    red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INIT);
     msg.visible = cursor_channel->cursor_visible;
     msg.position = cursor_channel->cursor_position;
     msg.trail_length = cursor_channel->cursor_trail_length;
     msg.trail_frequency = cursor_channel->cursor_trail_frequency;
 
-    cursor_fill(ccc, &msg.cursor, cursor_channel->item, &info);
+    cursor_fill(ccc, cursor_channel->item, &msg.cursor, base_marshaller);
     spice_marshall_msg_cursor_init(base_marshaller, &msg);
-    add_buf_from_info(base_marshaller, &info);
 }
 
 static void cursor_marshall(CursorChannelClient *ccc,
@@ -215,7 +206,6 @@ static void cursor_marshall(CursorChannelClient *ccc,
     RedChannelClient *rcc = RED_CHANNEL_CLIENT(ccc);
     CursorChannel *cursor_channel = CURSOR_CHANNEL(red_channel_client_get_channel(rcc));
     CursorItem *item = cursor_pipe_item->cursor_item;
-    RedPipeItem *pipe_item = &cursor_pipe_item->base;
     RedCursorCmd *cmd;
 
     spice_return_if_fail(cursor_channel);
@@ -225,7 +215,7 @@ static void cursor_marshall(CursorChannelClient *ccc,
     case QXL_CURSOR_MOVE:
         {
             SpiceMsgCursorMove cursor_move;
-            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_MOVE, pipe_item);
+            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_MOVE);
             cursor_move.position = cmd->u.position;
             spice_marshall_msg_cursor_move(m, &cursor_move);
             break;
@@ -233,25 +223,23 @@ static void cursor_marshall(CursorChannelClient *ccc,
     case QXL_CURSOR_SET:
         {
             SpiceMsgCursorSet cursor_set;
-            AddBufInfo info;
 
-            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_SET, pipe_item);
+            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_SET);
             cursor_set.position = cmd->u.set.position;
             cursor_set.visible = cursor_channel->cursor_visible;
 
-            cursor_fill(ccc, &cursor_set.cursor, item, &info);
+            cursor_fill(ccc, item, &cursor_set.cursor, m);
             spice_marshall_msg_cursor_set(m, &cursor_set);
-            add_buf_from_info(m, &info);
             break;
         }
     case QXL_CURSOR_HIDE:
-        red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_HIDE, pipe_item);
+        red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_HIDE);
         break;
     case QXL_CURSOR_TRAIL:
         {
             SpiceMsgCursorTrail cursor_trail;
 
-            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_TRAIL, pipe_item);
+            red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_TRAIL);
             cursor_trail.length = cmd->u.trail.length;
             cursor_trail.frequency = cmd->u.trail.frequency;
             spice_marshall_msg_cursor_trail(m, &cursor_trail);
@@ -268,7 +256,7 @@ static inline void red_marshall_inval(RedChannelClient *rcc,
 {
     SpiceMsgDisplayInvalOne inval_one;
 
-    red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INVAL_ONE, NULL);
+    red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INVAL_ONE);
     inval_one.id = cach_item->id;
 
     spice_marshall_msg_cursor_inval_one(base_marshaller, &inval_one);
@@ -292,7 +280,7 @@ static void cursor_channel_send_item(RedChannelClient *rcc, RedPipeItem *pipe_it
         break;
     case RED_PIPE_ITEM_TYPE_INVAL_CURSOR_CACHE:
         cursor_channel_client_reset_cursor_cache(rcc);
-        red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INVAL_ALL, NULL);
+        red_channel_client_init_send_data(rcc, SPICE_MSG_CURSOR_INVAL_ALL);
         break;
     default:
         spice_error("invalid pipe item type");
@@ -304,7 +292,7 @@ static void cursor_channel_send_item(RedChannelClient *rcc, RedPipeItem *pipe_it
 CursorChannel* cursor_channel_new(RedsState *server, QXLInstance *qxl,
                                   const SpiceCoreInterfaceInternal *core)
 {
-    spice_info("create cursor channel");
+    spice_debug("create cursor channel");
     return g_object_new(TYPE_CURSOR_CHANNEL,
                         "spice-server", server,
                         "core-interface", core,
@@ -319,7 +307,7 @@ CursorChannel* cursor_channel_new(RedsState *server, QXLInstance *qxl,
 void cursor_channel_process_cmd(CursorChannel *cursor, RedCursorCmd *cursor_cmd)
 {
     CursorItem *cursor_item;
-    int cursor_show = FALSE;
+    bool cursor_show = false;
     QXLInstance *qxl;
 
     spice_return_if_fail(cursor);
@@ -330,16 +318,16 @@ void cursor_channel_process_cmd(CursorChannel *cursor, RedCursorCmd *cursor_cmd)
 
     switch (cursor_cmd->type) {
     case QXL_CURSOR_SET:
-        cursor->cursor_visible = cursor_cmd->u.set.visible;
-        cursor_set_item(cursor, cursor_item);
+        cursor->cursor_visible = !!cursor_cmd->u.set.visible;
+        cursor_channel_set_item(cursor, cursor_item);
         break;
     case QXL_CURSOR_MOVE:
         cursor_show = !cursor->cursor_visible;
-        cursor->cursor_visible = TRUE;
+        cursor->cursor_visible = true;
         cursor->cursor_position = cursor_cmd->u.position;
         break;
     case QXL_CURSOR_HIDE:
-        cursor->cursor_visible = FALSE;
+        cursor->cursor_visible = false;
         break;
     case QXL_CURSOR_TRAIL:
         cursor->cursor_trail_length = cursor_cmd->u.trail.length;
@@ -367,8 +355,8 @@ void cursor_channel_reset(CursorChannel *cursor)
 
     spice_return_if_fail(cursor);
 
-    cursor_set_item(cursor, NULL);
-    cursor->cursor_visible = TRUE;
+    cursor_channel_set_item(cursor, NULL);
+    cursor->cursor_visible = true;
     cursor->cursor_position.x = cursor->cursor_position.y = 0;
     cursor->cursor_trail_length = cursor->cursor_trail_frequency = 0;
 
@@ -416,18 +404,16 @@ void cursor_channel_set_mouse_mode(CursorChannel *cursor, uint32_t mode)
 
 void cursor_channel_connect(CursorChannel *cursor, RedClient *client, RedsStream *stream,
                             int migrate,
-                            uint32_t *common_caps, int num_common_caps,
-                            uint32_t *caps, int num_caps)
+                            RedChannelCapabilities *caps)
 {
     CursorChannelClient *ccc;
 
     spice_return_if_fail(cursor != NULL);
 
-    spice_info("add cursor channel client");
+    spice_debug("add cursor channel client");
     ccc = cursor_channel_client_new(cursor, client, stream,
                                     migrate,
-                                    common_caps, num_common_caps,
-                                    caps, num_caps);
+                                    caps);
     spice_return_if_fail(ccc != NULL);
 
     RedChannelClient *rcc = RED_CHANNEL_CLIENT(ccc);
@@ -437,13 +423,28 @@ void cursor_channel_connect(CursorChannel *cursor, RedClient *client, RedsStream
     cursor_channel_init_client(cursor, ccc);
 }
 
+static void
+cursor_channel_finalize(GObject *object)
+{
+    CursorChannel *self = CURSOR_CHANNEL(object);
+
+    if (self->item) {
+        cursor_item_unref(self->item);
+    }
+
+    G_OBJECT_CLASS(cursor_channel_parent_class)->finalize(object);
+}
+
 static void
 cursor_channel_class_init(CursorChannelClass *klass)
 {
+    GObjectClass *object_class = G_OBJECT_CLASS(klass);
     RedChannelClass *channel_class = RED_CHANNEL_CLASS(klass);
 
+    object_class->finalize = cursor_channel_finalize;
+
     channel_class->parser = spice_get_client_channel_parser(SPICE_CHANNEL_CURSOR, NULL);
-    channel_class->handle_parsed = red_channel_client_handle_message;
+    channel_class->handle_message = red_channel_client_handle_message;
 
     channel_class->on_disconnect =  cursor_channel_client_on_disconnect;
     channel_class->send_item = cursor_channel_send_item;
@@ -452,6 +453,6 @@ cursor_channel_class_init(CursorChannelClass *klass)
 static void
 cursor_channel_init(CursorChannel *self)
 {
-    self->cursor_visible = TRUE;
+    self->cursor_visible = true;
     self->mouse_mode = SPICE_MOUSE_MODE_SERVER;
 }
