@@ -21,10 +21,12 @@
 #endif
 
 #include <inttypes.h>
+#include <pthread.h>
 
 #include <gst/gst.h>
 #include <gst/app/gstappsrc.h>
 #include <gst/app/gstappsink.h>
+#include <gst/video/video.h>
 
 #include "red-common.h"
 #include "video-encoder.h"
@@ -40,15 +42,29 @@
 
 typedef struct {
     SpiceBitmapFmt spice_format;
-    const char *format;
     uint32_t bpp;
+#ifndef HAVE_GSTREAMER_0_10
+    char format[8];
+    GstVideoFormat gst_format;
+#else
     uint32_t depth;
     uint32_t endianness;
     uint32_t blue_mask;
     uint32_t green_mask;
     uint32_t red_mask;
+#endif
 } SpiceFormatForGStreamer;
 
+#ifndef HAVE_GSTREAMER_0_10
+#define FMT_DESC(spice_format, bpp, format, gst_format, depth, endianness, \
+                 blue_mask, green_mask, red_mask) \
+    { spice_format, bpp, format, gst_format }
+#else
+#define FMT_DESC(spice_format, bpp, format, gst_format, depth, endianness, \
+                 blue_mask, green_mask, red_mask) \
+    { spice_format, bpp, depth, endianness, blue_mask, green_mask, red_mask }
+#endif
+
 typedef struct SpiceGstVideoBuffer {
     VideoBuffer base;
     GstBuffer *gst_buffer;
@@ -118,8 +134,8 @@ typedef struct SpiceGstEncoder {
     uint32_t set_pipeline;
 
     /* Output buffer */
-    GMutex outbuf_mutex;
-    GCond outbuf_cond;
+    pthread_mutex_t outbuf_mutex;
+    pthread_cond_t outbuf_cond;
     VideoBuffer *outbuf;
 
     /* The video bit rate. */
@@ -182,7 +198,8 @@ typedef struct SpiceGstEncoder {
     /* The bit rate control is performed using a virtual buffer to allow
      * short term variations: bursts are allowed until the virtual buffer is
      * full. Then frames are dropped to limit the bit rate. VBUFFER_SIZE
-     * defines the size of the virtual buffer in milliseconds worth of data.
+     * defines the size of the virtual buffer in milliseconds worth of data
+     * while vbuffer_size holds the limit in bytes for the current bit rate.
      */
 #   define SPICE_GST_VBUFFER_SIZE 300
 
@@ -330,11 +347,6 @@ static uint32_t get_network_latency(SpiceGstEncoder *encoder)
         encoder->cbs.get_roundtrip_ms(encoder->cbs.opaque) / 2 : 0;
 }
 
-static inline int rate_control_is_active(SpiceGstEncoder* encoder)
-{
-    return encoder->cbs.get_roundtrip_ms != NULL;
-}
-
 static void set_pipeline_changes(SpiceGstEncoder *encoder, uint32_t flags)
 {
     encoder->set_pipeline |= flags;
@@ -434,19 +446,20 @@ static uint64_t get_period_bit_rate(SpiceGstEncoder *encoder, uint32_t from,
             sum += encoder->history[index].size;
             return (sum - 1) * 8 * MSEC_PER_SEC / (last_mm_time - from);
 
-        } else if (index == encoder->history_first) {
+        } else if (sum > 0) {
+            sum += encoder->history[index].size;
+
+        } else {
+            last_mm_time = encoder->history[index].mm_time;
+        }
+
+        if (index == encoder->history_first) {
             /* This period is outside the recorded history */
             spice_debug("period (%u-%u) outside known history (%u-%u)",
                         from, to,
                         encoder->history[encoder->history_first].mm_time,
                         encoder->history[encoder->history_last].mm_time);
            return 0;
-
-        } else if (sum > 0) {
-            sum += encoder->history[index].size;
-
-        } else {
-            last_mm_time = encoder->history[index].mm_time;
         }
         index = (index ? index : SPICE_GST_HISTORY_SIZE) - 1;
     }
@@ -508,7 +521,7 @@ static uint32_t get_min_playback_delay(SpiceGstEncoder *encoder)
      * an I frame) and an average frame. This also takes into account the
      * frames dropped by the encoder bit rate control.
      */
-    uint32_t size = get_maximum_frame_size(encoder) + get_average_frame_size(encoder);
+    uint64_t size = get_maximum_frame_size(encoder) + get_average_frame_size(encoder);
     uint32_t send_time = MSEC_PER_SEC * size * 8 / encoder->bit_rate;
 
     /* Also factor in the network latency with a margin for jitter. */
@@ -592,7 +605,7 @@ static void set_bit_rate(SpiceGstEncoder *encoder, uint64_t bit_rate)
     }
     if (bit_rate < SPICE_GST_MIN_BITRATE) {
         /* Don't let the bit rate go too low... */
-        encoder->bit_rate = SPICE_GST_MIN_BITRATE;
+        bit_rate = SPICE_GST_MIN_BITRATE;
     } else if (bit_rate > encoder->bit_rate) {
         /* or too high */
         bit_rate = MIN(bit_rate, get_bit_rate_cap(encoder));
@@ -746,31 +759,38 @@ static inline void server_increase_bit_rate(SpiceGstEncoder *encoder,
 
 /* ---------- GStreamer pipeline ---------- */
 
+/* See GStreamer's part-mediatype-video-raw.txt and
+ * section-types-definitions.html documents.
+ */
+static const SpiceFormatForGStreamer format_map[] =  {
+    /* First item is invalid.
+     * It's located first so the loop catch invalid values.
+     */
+    FMT_DESC(SPICE_BITMAP_FMT_INVALID, 0, "", GST_VIDEO_FORMAT_UNKNOWN, 0, 0, 0, 0, 0),
+    FMT_DESC(SPICE_BITMAP_FMT_RGBA, 32, "BGRA", GST_VIDEO_FORMAT_BGRA, 24, 4321, 0xff000000, 0xff0000, 0xff00),
+    FMT_DESC(SPICE_BITMAP_FMT_16BIT, 16, "RGB15", GST_VIDEO_FORMAT_RGB15, 15, 4321, 0x001f, 0x03E0, 0x7C00),
+    /* TODO: Test the other formats under GStreamer 0.10*/
+    FMT_DESC(SPICE_BITMAP_FMT_32BIT, 32, "BGRx", GST_VIDEO_FORMAT_BGRx, 24, 4321, 0xff000000, 0xff0000, 0xff00),
+    FMT_DESC(SPICE_BITMAP_FMT_24BIT, 24, "BGR", GST_VIDEO_FORMAT_BGR, 24, 4321, 0xff0000, 0xff00, 0xff),
+};
+#define GSTREAMER_FORMAT_INVALID (&format_map[0])
+
 /* A helper for spice_gst_encoder_encode_frame() */
 static const SpiceFormatForGStreamer *map_format(SpiceBitmapFmt format)
 {
-    /* See GStreamer's part-mediatype-video-raw.txt and
-     * section-types-definitions.html documents.
-     */
-    static const SpiceFormatForGStreamer format_map[] =  {
-        {SPICE_BITMAP_FMT_RGBA, "BGRA", 32, 24, 4321, 0xff000000, 0xff0000, 0xff00},
-        {SPICE_BITMAP_FMT_16BIT, "RGB15", 16, 15, 4321, 0x001f, 0x03E0, 0x7C00},
-        /* TODO: Test the other formats */
-        {SPICE_BITMAP_FMT_32BIT, "BGRx", 32, 24, 4321, 0xff000000, 0xff0000, 0xff00},
-        {SPICE_BITMAP_FMT_24BIT, "BGR", 24, 24, 4321, 0xff0000, 0xff00, 0xff},
-    };
-
     int i;
     for (i = 0; i < G_N_ELEMENTS(format_map); i++) {
         if (format_map[i].spice_format == format) {
-            if (i > 1) {
+#ifdef HAVE_GSTREAMER_0_10
+            if (i > 2) {
                 spice_warning("The %d format has not been tested yet", format);
             }
+#endif
             return &format_map[i];
         }
     }
 
-    return NULL;
+    return GSTREAMER_FORMAT_INVALID;
 }
 
 static void set_appsrc_caps(SpiceGstEncoder *encoder)
@@ -815,10 +835,10 @@ static GstBusSyncReply handle_pipeline_message(GstBus *bus, GstMessage *msg, gpo
         g_clear_error(&err);
 
         /* Unblock the main thread */
-        g_mutex_lock(&encoder->outbuf_mutex);
+        pthread_mutex_lock(&encoder->outbuf_mutex);
         encoder->outbuf = (VideoBuffer*)create_gst_video_buffer();
-        g_cond_signal(&encoder->outbuf_cond);
-        g_mutex_unlock(&encoder->outbuf_mutex);
+        pthread_cond_signal(&encoder->outbuf_cond);
+        pthread_mutex_unlock(&encoder->outbuf_mutex);
     }
     return GST_BUS_PASS;
 }
@@ -848,10 +868,10 @@ static GstFlowReturn new_sample(GstAppSink *gstappsink, gpointer video_encoder)
 #endif
 
     /* Notify the main thread that the output buffer is ready */
-    g_mutex_lock(&encoder->outbuf_mutex);
+    pthread_mutex_lock(&encoder->outbuf_mutex);
     encoder->outbuf = (VideoBuffer*)outbuf;
-    g_cond_signal(&encoder->outbuf_cond);
-    g_mutex_unlock(&encoder->outbuf_mutex);
+    pthread_cond_signal(&encoder->outbuf_cond);
+    pthread_mutex_unlock(&encoder->outbuf_mutex);
 
     return GST_FLOW_OK;
 }
@@ -870,6 +890,8 @@ static const gchar* get_gst_codec_name(SpiceGstEncoder *encoder)
         return "vp8enc";
     case SPICE_VIDEO_CODEC_TYPE_H264:
         return "x264enc";
+    case SPICE_VIDEO_CODEC_TYPE_VP9:
+        return "vp9enc";
     default:
         /* gstreamer_encoder_new() should have rejected this codec type */
         spice_warning("unsupported codec type %d", encoder->base.codec_type);
@@ -899,6 +921,7 @@ static gboolean create_pipeline(SpiceGstEncoder *encoder)
         gstenc_opts = g_strdup("max-threads=1");
 #endif
         break;
+    case SPICE_VIDEO_CODEC_TYPE_VP9:
     case SPICE_VIDEO_CODEC_TYPE_VP8: {
         /* See http://www.webmproject.org/docs/encoder-parameters/
          * - Set mode/end-usage to get a constant bitrate to help with
@@ -926,11 +949,14 @@ static gboolean create_pipeline(SpiceGstEncoder *encoder)
     case SPICE_VIDEO_CODEC_TYPE_H264:
         /* - Set tune and sliced-threads to ensure a zero-frame latency
          * - qp-min ensures the bitrate does not get needlessly high.
+         * - qp-max ensures the compression does not go so high that the video
+         *   is unrecognizable. When that threshold is reached it is better to
+         *   drop frames to lower the bit rate further.
          * - Set speed-preset to get realtime speed.
          * - Set intra-refresh to get more uniform compressed frame sizes,
          *   thus helping with streaming.
          */
-        gstenc_opts = g_strdup("byte-stream=true aud=true qp-min=15 tune=4 sliced-threads=true speed-preset=ultrafast intra-refresh=true");
+        gstenc_opts = g_strdup("byte-stream=true aud=true qp-min=15 qp-max=35 tune=4 sliced-threads=true speed-preset=ultrafast intra-refresh=true");
         break;
     default:
         /* gstreamer_encoder_new() should have rejected this codec type */
@@ -1114,36 +1140,6 @@ static int is_chunk_stride_aligned(const SpiceBitmap *bitmap, uint32_t index)
     return TRUE;
 }
 
-/* A helper for push_raw_frame() */
-static inline int line_copy(SpiceGstEncoder *encoder, const SpiceBitmap *bitmap,
-                            uint32_t chunk_offset, uint32_t stream_stride,
-                            uint32_t height, uint8_t *buffer)
-{
-     uint8_t *dst = buffer;
-     SpiceChunks *chunks = bitmap->data;
-     uint32_t chunk_index = 0;
-     for (int l = 0; l < height; l++) {
-         /* We may have to move forward by more than one chunk the first
-          * time around. This also protects us against 0-byte chunks.
-          */
-         while (chunk_offset >= chunks->chunk[chunk_index].len) {
-             if (!is_chunk_stride_aligned(bitmap, chunk_index)) {
-                 return FALSE;
-             }
-             chunk_offset -= chunks->chunk[chunk_index].len;
-             chunk_index++;
-         }
-
-         /* Copy the line */
-         uint8_t *src = chunks->chunk[chunk_index].data + chunk_offset;
-         memcpy(dst, src, stream_stride);
-         dst += stream_stride;
-         chunk_offset += bitmap->stride;
-     }
-     spice_return_val_if_fail(dst - buffer == stream_stride * height, FALSE);
-     return TRUE;
-}
-
 #ifdef DO_ZERO_COPY
 typedef struct {
     gint refs;
@@ -1239,8 +1235,39 @@ static void clear_zero_copy_queue(SpiceGstEncoder *encoder, gboolean unref_queue
 {
     /* Nothing to do */
 }
+
 #endif
 
+/* A helper for push_raw_frame() */
+static inline int line_copy(SpiceGstEncoder *encoder, const SpiceBitmap *bitmap,
+                            uint32_t chunk_offset, uint32_t stream_stride,
+                            uint32_t height, uint8_t *buffer)
+{
+     uint8_t *dst = buffer;
+     SpiceChunks *chunks = bitmap->data;
+     uint32_t chunk_index = 0;
+     for (int l = 0; l < height; l++) {
+         /* We may have to move forward by more than one chunk the first
+          * time around. This also protects us against 0-byte chunks.
+          */
+         while (chunk_offset >= chunks->chunk[chunk_index].len) {
+             if (!is_chunk_stride_aligned(bitmap, chunk_index)) {
+                 return FALSE;
+             }
+             chunk_offset -= chunks->chunk[chunk_index].len;
+             chunk_index++;
+         }
+
+         /* Copy the line */
+         uint8_t *src = chunks->chunk[chunk_index].data + chunk_offset;
+         memcpy(dst, src, MIN(stream_stride, bitmap->stride));
+         dst += stream_stride;
+         chunk_offset += bitmap->stride;
+     }
+     spice_return_val_if_fail(dst - buffer == stream_stride * height, FALSE);
+     return TRUE;
+}
+
 /* A helper for push_raw_frame() */
 static inline int chunk_copy(SpiceGstEncoder *encoder, const SpiceBitmap *bitmap,
                              uint32_t chunk_index, uint32_t chunk_offset,
@@ -1323,7 +1350,8 @@ static int push_raw_frame(SpiceGstEncoder *encoder,
                           gpointer bitmap_opaque)
 {
     uint32_t height = src->bottom - src->top;
-    uint32_t stream_stride = (src->right - src->left) * encoder->format->bpp / 8;
+    // GStreamer require the stream to be 4 bytes aligned
+    uint32_t stream_stride = GST_ROUND_UP_4((src->right - src->left) * encoder->format->bpp / 8);
     uint32_t len = stream_stride * height;
     GstBuffer *buffer = gst_buffer_new();
     /* TODO Use GST_MAP_INFO_INIT once GStreamer 1.4.5 is no longer relevant */
@@ -1399,13 +1427,13 @@ static int push_raw_frame(SpiceGstEncoder *encoder,
 static int pull_compressed_buffer(SpiceGstEncoder *encoder,
                                   VideoBuffer **outbuf)
 {
-    g_mutex_lock(&encoder->outbuf_mutex);
+    pthread_mutex_lock(&encoder->outbuf_mutex);
     while (!encoder->outbuf) {
-        g_cond_wait(&encoder->outbuf_cond, &encoder->outbuf_mutex);
+        pthread_cond_wait(&encoder->outbuf_cond, &encoder->outbuf_mutex);
     }
     *outbuf = encoder->outbuf;
     encoder->outbuf = NULL;
-    g_mutex_unlock(&encoder->outbuf_mutex);
+    pthread_mutex_unlock(&encoder->outbuf_mutex);
 
     if ((*outbuf)->data) {
         return VIDEO_ENCODER_FRAME_ENCODE_DONE;
@@ -1425,8 +1453,8 @@ static void spice_gst_encoder_destroy(VideoEncoder *video_encoder)
     SpiceGstEncoder *encoder = (SpiceGstEncoder*)video_encoder;
 
     free_pipeline(encoder);
-    g_mutex_clear(&encoder->outbuf_mutex);
-    g_cond_clear(&encoder->outbuf_cond);
+    pthread_mutex_destroy(&encoder->outbuf_mutex);
+    pthread_cond_destroy(&encoder->outbuf_cond);
 
     /* Unref any lingering bitmap opaque structures from past frames */
     clear_zero_copy_queue(encoder, TRUE);
@@ -1456,7 +1484,7 @@ static int spice_gst_encoder_encode_frame(VideoEncoder *video_encoder,
                     encoder->width, width, encoder->height, height,
                     encoder->spice_format, bitmap->format);
         encoder->format = map_format(bitmap->format);
-        if (!encoder->format) {
+        if (encoder->format == GSTREAMER_FORMAT_INVALID) {
             spice_warning("unable to map format type %d", bitmap->format);
             encoder->errors = 4;
             return VIDEO_ENCODER_FRAME_UNSUPPORTED;
@@ -1489,9 +1517,8 @@ static int spice_gst_encoder_encode_frame(VideoEncoder *video_encoder,
         return VIDEO_ENCODER_FRAME_UNSUPPORTED;
     }
 
-    if (rate_control_is_active(encoder) &&
-        (handle_server_drops(encoder, frame_mm_time) ||
-         frame_mm_time < encoder->next_frame_mm_time)) {
+    if (handle_server_drops(encoder, frame_mm_time) ||
+        frame_mm_time < encoder->next_frame_mm_time) {
         /* Drop the frame to limit the outgoing bit rate. */
         return VIDEO_ENCODER_FRAME_DROP;
     }
@@ -1662,7 +1689,7 @@ static void spice_gst_encoder_get_stats(VideoEncoder *video_encoder,
                                         VideoEncoderStats *stats)
 {
     SpiceGstEncoder *encoder = (SpiceGstEncoder*)video_encoder;
-    uint64_t raw_bit_rate = encoder->width * encoder->height * (encoder->format ? encoder->format->bpp : 0) * get_source_fps(encoder);
+    uint64_t raw_bit_rate = encoder->width * encoder->height * encoder->format->bpp * get_source_fps(encoder);
 
     spice_return_if_fail(stats != NULL);
     stats->starting_bit_rate = encoder->starting_bit_rate;
@@ -1681,9 +1708,10 @@ VideoEncoder *gstreamer_encoder_new(SpiceVideoCodecType codec_type,
                                     bitmap_ref_t bitmap_ref,
                                     bitmap_unref_t bitmap_unref)
 {
-    spice_return_val_if_fail(SPICE_GST_FRAME_STATISTICS_COUNT <= SPICE_GST_HISTORY_SIZE, NULL);
+    SPICE_VERIFY(SPICE_GST_FRAME_STATISTICS_COUNT <= SPICE_GST_HISTORY_SIZE);
     spice_return_val_if_fail(codec_type == SPICE_VIDEO_CODEC_TYPE_MJPEG ||
                              codec_type == SPICE_VIDEO_CODEC_TYPE_VP8 ||
+                             codec_type == SPICE_VIDEO_CODEC_TYPE_VP9 ||
                              codec_type == SPICE_VIDEO_CODEC_TYPE_H264, NULL);
 
     GError *err = NULL;
@@ -1705,19 +1733,20 @@ VideoEncoder *gstreamer_encoder_new(SpiceVideoCodecType codec_type,
     encoder->unused_bitmap_opaques = g_async_queue_new();
 #endif
 
-    if (cbs) {
-        encoder->cbs = *cbs;
-    }
     encoder->starting_bit_rate = starting_bit_rate;
+    encoder->cbs = *cbs;
     encoder->bitmap_ref = bitmap_ref;
     encoder->bitmap_unref = bitmap_unref;
-    g_mutex_init(&encoder->outbuf_mutex);
-    g_cond_init(&encoder->outbuf_cond);
+    encoder->format = GSTREAMER_FORMAT_INVALID;
+    pthread_mutex_init(&encoder->outbuf_mutex, NULL);
+    pthread_cond_init(&encoder->outbuf_cond, NULL);
 
     /* All the other fields are initialized to zero by spice_new0(). */
 
     if (!create_pipeline(encoder)) {
         /* Some GStreamer dependency is probably missing */
+        pthread_cond_destroy(&encoder->outbuf_cond);
+        pthread_mutex_destroy(&encoder->outbuf_mutex);
         free(encoder);
         encoder = NULL;
     }
