@@ -23,7 +23,7 @@
 
 #include "image-encoders.h"
 #include "spice-bitmap-utils.h"
-#include "red-worker.h" // red_drawable_unref
+#include "red-parse-qxl.h" // red_drawable_unref
 #include "pixmap-cache.h" // MAX_CACHE_CLIENTS
 
 #define ZLIB_DEFAULT_COMPRESSION_LEVEL 3
@@ -36,7 +36,6 @@ typedef struct RedGlzDrawable RedGlzDrawable;
 typedef struct GlzDrawableInstanceItem GlzDrawableInstanceItem;
 
 struct GlzSharedDictionary {
-    RingItem base;
     GlzEncDictContext *dict;
     uint32_t refs;
     uint8_t id;
@@ -694,23 +693,21 @@ static GlzSharedDictionary *glz_shared_dictionary_new(RedClient *client, uint8_t
     shared_dict->refs = 1;
     shared_dict->migrate_freeze = FALSE;
     shared_dict->client = client;
-    ring_item_init(&shared_dict->base);
     pthread_rwlock_init(&shared_dict->encode_lock, NULL);
 
     return shared_dict;
 }
 
 static pthread_mutex_t glz_dictionary_list_lock = PTHREAD_MUTEX_INITIALIZER;
-static Ring glz_dictionary_list = {&glz_dictionary_list, &glz_dictionary_list};
+static GList *glz_dictionary_list;
 
 static GlzSharedDictionary *find_glz_dictionary(RedClient *client, uint8_t dict_id)
 {
-    RingItem *now;
+    GList *l;
     GlzSharedDictionary *ret = NULL;
 
-    now = &glz_dictionary_list;
-    while ((now = ring_next(&glz_dictionary_list, now))) {
-        GlzSharedDictionary *dict = SPICE_UPCAST(GlzSharedDictionary, now);
+    for (l = glz_dictionary_list; l != NULL; l = l->next) {
+        GlzSharedDictionary *dict = l->data;
         if ((dict->client == client) && (dict->id == dict_id)) {
             ret = dict;
             break;
@@ -726,7 +723,7 @@ static GlzSharedDictionary *create_glz_dictionary(ImageEncoders *enc,
                                                   RedClient *client,
                                                   uint8_t id, int window_size)
 {
-    spice_info("Lz Window %d Size=%d", id, window_size);
+    spice_debug("Lz Window %d Size=%d", id, window_size);
 
     GlzEncDictContext *glz_dict =
         glz_enc_dictionary_create(window_size, MAX_LZ_ENCODERS, &enc->glz_data.usr);
@@ -749,7 +746,9 @@ gboolean image_encoders_get_glz_dictionary(ImageEncoders *enc,
         shared_dict->refs++;
     } else {
         shared_dict = create_glz_dictionary(enc, client, id, window_size);
-        ring_add(&glz_dictionary_list, &shared_dict->base);
+        if (shared_dict != NULL) {
+            glz_dictionary_list = g_list_prepend(glz_dictionary_list, shared_dict);
+        }
     }
 
     pthread_mutex_unlock(&glz_dictionary_list_lock);
@@ -785,7 +784,9 @@ gboolean image_encoders_restore_glz_dictionary(ImageEncoders *enc,
         shared_dict->refs++;
     } else {
         shared_dict = restore_glz_dictionary(enc, client, id, restore_data);
-        ring_add(&glz_dictionary_list, &shared_dict->base);
+        if(shared_dict != NULL) {
+            glz_dictionary_list = g_list_prepend(glz_dictionary_list, shared_dict);
+        }
     }
 
     pthread_mutex_unlock(&glz_dictionary_list_lock);
@@ -819,15 +820,15 @@ static void image_encoders_release_glz(ImageEncoders *enc)
         pthread_mutex_unlock(&glz_dictionary_list_lock);
         return;
     }
-    ring_remove(&shared_dict->base);
+    glz_dictionary_list = g_list_remove(glz_dictionary_list, shared_dict);
     pthread_mutex_unlock(&glz_dictionary_list_lock);
     glz_enc_dictionary_destroy(shared_dict->dict, &enc->glz_data.usr);
     pthread_rwlock_destroy(&shared_dict->encode_lock);
     free(shared_dict);
 }
 
-int image_encoders_compress_quic(ImageEncoders *enc, SpiceImage *dest,
-                                 SpiceBitmap *src, compress_send_data_t* o_comp_data)
+bool image_encoders_compress_quic(ImageEncoders *enc, SpiceImage *dest,
+                                  SpiceBitmap *src, compress_send_data_t* o_comp_data)
 {
     QuicData *quic_data = &enc->quic_data;
     QuicContext *quic = enc->quic;
@@ -837,7 +838,7 @@ int image_encoders_compress_quic(ImageEncoders *enc, SpiceImage *dest,
     stat_start_time_init(&start_time, &enc->shared_data->quic_stat);
 
 #ifdef COMPRESS_DEBUG
-    spice_info("QUIC compress");
+    spice_debug("QUIC compress");
 #endif
 
     switch (src->format) {
@@ -913,9 +914,9 @@ static const LzImageType bitmap_fmt_to_lz_image_type[] = {
     LZ_IMAGE_TYPE_A8
 };
 
-int image_encoders_compress_lz(ImageEncoders *enc,
-                               SpiceImage *dest, SpiceBitmap *src,
-                               compress_send_data_t* o_comp_data)
+bool image_encoders_compress_lz(ImageEncoders *enc,
+                                SpiceImage *dest, SpiceBitmap *src,
+                                compress_send_data_t* o_comp_data)
 {
     LzData *lz_data = &enc->lz_data;
     LzContext *lz = enc->lz;
@@ -926,7 +927,7 @@ int image_encoders_compress_lz(ImageEncoders *enc,
     stat_start_time_init(&start_time, &enc->shared_data->lz_stat);
 
 #ifdef COMPRESS_DEBUG
-    spice_info("LZ LOCAL compress");
+    spice_debug("LZ LOCAL compress");
 #endif
 
     encoder_data_init(&lz_data->data);
@@ -978,8 +979,8 @@ int image_encoders_compress_lz(ImageEncoders *enc,
     return TRUE;
 }
 
-int image_encoders_compress_jpeg(ImageEncoders *enc, SpiceImage *dest,
-                                 SpiceBitmap *src, compress_send_data_t* o_comp_data)
+bool image_encoders_compress_jpeg(ImageEncoders *enc, SpiceImage *dest,
+                                  SpiceBitmap *src, compress_send_data_t* o_comp_data)
 {
     JpegData *jpeg_data = &enc->jpeg_data;
     LzData *lz_data = &enc->lz_data;
@@ -997,7 +998,7 @@ int image_encoders_compress_jpeg(ImageEncoders *enc, SpiceImage *dest,
     stat_start_time_init(&start_time, &enc->shared_data->jpeg_alpha_stat);
 
 #ifdef COMPRESS_DEBUG
-    spice_info("JPEG compress");
+    spice_debug("JPEG compress");
 #endif
 
     switch (src->format) {
@@ -1104,8 +1105,8 @@ int image_encoders_compress_jpeg(ImageEncoders *enc, SpiceImage *dest,
 }
 
 #ifdef USE_LZ4
-int image_encoders_compress_lz4(ImageEncoders *enc, SpiceImage *dest,
-                                SpiceBitmap *src, compress_send_data_t* o_comp_data)
+bool image_encoders_compress_lz4(ImageEncoders *enc, SpiceImage *dest,
+                                 SpiceBitmap *src, compress_send_data_t* o_comp_data)
 {
     Lz4Data *lz4_data = &enc->lz4_data;
     Lz4EncoderContext *lz4 = enc->lz4;
@@ -1114,7 +1115,7 @@ int image_encoders_compress_lz4(ImageEncoders *enc, SpiceImage *dest,
     stat_start_time_init(&start_time, &enc->shared_data->lz4_stat);
 
 #ifdef COMPRESS_DEBUG
-    spice_info("LZ4 compress");
+    spice_debug("LZ4 compress");
 #endif
 
     encoder_data_init(&lz4_data->data);
@@ -1207,12 +1208,12 @@ static GlzDrawableInstanceItem *add_glz_drawable_instance(RedGlzDrawable *glz_dr
 
 #define MIN_GLZ_SIZE_FOR_ZLIB 100
 
-int image_encoders_compress_glz(ImageEncoders *enc,
-                                SpiceImage *dest, SpiceBitmap *src,
-                                RedDrawable *red_drawable,
-                                GlzImageRetention *glz_retention,
-                                compress_send_data_t* o_comp_data,
-                                gboolean enable_zlib_glz_wrap)
+bool image_encoders_compress_glz(ImageEncoders *enc,
+                                 SpiceImage *dest, SpiceBitmap *src,
+                                 RedDrawable *red_drawable,
+                                 GlzImageRetention *glz_retention,
+                                 compress_send_data_t* o_comp_data,
+                                 gboolean enable_zlib_glz_wrap)
 {
     stat_start_time_t start_time;
     stat_start_time_init(&start_time, &enc->shared_data->zlib_glz_stat);
@@ -1226,7 +1227,7 @@ int image_encoders_compress_glz(ImageEncoders *enc,
     int zlib_size;
 
 #ifdef COMPRESS_DEBUG
-    spice_info("LZ global compress fmt=%d", src->format);
+    spice_debug("LZ global compress fmt=%d", src->format);
 #endif
 
     if ((src->x * src->y) >= glz_enc_dictionary_get_size(enc->glz_dict->dict)) {
