@@ -15,8 +15,8 @@
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
 
-#ifndef __MAIN_CHANNEL_H__
-#define __MAIN_CHANNEL_H__
+#ifndef MAIN_CHANNEL_H_
+#define MAIN_CHANNEL_H_
 
 #include <stdint.h>
 #include <glib-object.h>
@@ -59,15 +59,12 @@ MainChannel *main_channel_new(RedsState *reds);
 RedClient *main_channel_get_client_by_link_id(MainChannel *main_chan, uint32_t link_id);
 /* This is a 'clone' from the reds.h Channel.link callback to allow passing link_id */
 MainChannelClient *main_channel_link(MainChannel *, RedClient *client,
-     RedsStream *stream, uint32_t link_id, int migration, int num_common_caps,
-     uint32_t *common_caps, int num_caps, uint32_t *caps);
-void main_channel_close(MainChannel *main_chan); // not destroy, just socket close
+     RedsStream *stream, uint32_t link_id, int migration,
+     RedChannelCapabilities *caps);
 void main_channel_push_mouse_mode(MainChannel *main_chan, int current_mode, int is_client_mouse_allowed);
 void main_channel_push_agent_connected(MainChannel *main_chan);
 void main_channel_push_agent_disconnected(MainChannel *main_chan);
 void main_channel_push_multi_media_time(MainChannel *main_chan, int time);
-int main_channel_getsockname(MainChannel *main_chan, struct sockaddr *sa, socklen_t *salen);
-int main_channel_getpeername(MainChannel *main_chan, struct sockaddr *sa, socklen_t *salen);
 
 int main_channel_is_connected(MainChannel *main_chan);
 
@@ -89,4 +86,4 @@ void main_channel_on_migrate_connected(MainChannel *main_channel,
 
 G_END_DECLS
 
-#endif
+#endif /* MAIN_CHANNEL_H_ */
