@@ -36,6 +36,7 @@
 #include "reds.h"
 #include "dispatcher.h"
 #include "red-parse-qxl.h"
+#include "red-channel-client.h"
 
 #include "red-qxl.h"
 
@@ -57,6 +58,7 @@ struct QXLState {
     QXLDevSurfaceCreate surface_create;
     unsigned int max_monitors;
     RedsState *reds;
+    RedWorker *worker;
 
     pthread_mutex_t scanout_mutex;
     SpiceMsgDisplayGlScanoutUnix scanout;
@@ -74,24 +76,17 @@ int red_qxl_check_qxl_version(QXLInstance *qxl, int major, int minor)
 
 static void red_qxl_set_display_peer(RedChannel *channel, RedClient *client,
                                      RedsStream *stream, int migration,
-                                     int num_common_caps, uint32_t *common_caps, int num_caps,
-                                     uint32_t *caps)
+                                     RedChannelCapabilities *caps)
 {
     RedWorkerMessageDisplayConnect payload = {0,};
     Dispatcher *dispatcher;
 
     spice_debug("%s", "");
-    dispatcher = (Dispatcher *)channel->data;
+    dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
     payload.client = client;
     payload.stream = stream;
     payload.migration = migration;
-    payload.num_common_caps = num_common_caps;
-    payload.common_caps = spice_malloc(sizeof(uint32_t)*num_common_caps);
-    payload.num_caps = num_caps;
-    payload.caps = spice_malloc(sizeof(uint32_t)*num_caps);
-
-    memcpy(payload.common_caps, common_caps, sizeof(uint32_t)*num_common_caps);
-    memcpy(payload.caps, caps, sizeof(uint32_t)*num_caps);
+    red_channel_capabilities_init(&payload.caps, caps);
 
     dispatcher_send_message(dispatcher,
                             RED_WORKER_MESSAGE_DISPLAY_CONNECT,
@@ -102,12 +97,13 @@ static void red_qxl_disconnect_display_peer(RedChannelClient *rcc)
 {
     RedWorkerMessageDisplayDisconnect payload;
     Dispatcher *dispatcher;
+    RedChannel *channel = red_channel_client_get_channel(rcc);
 
-    if (!rcc->channel) {
+    if (!channel) {
         return;
     }
 
-    dispatcher = (Dispatcher *)rcc->channel->data;
+    dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
 
     spice_printerr("");
     payload.rcc = rcc;
@@ -123,11 +119,15 @@ static void red_qxl_display_migrate(RedChannelClient *rcc)
 {
     RedWorkerMessageDisplayMigrate payload;
     Dispatcher *dispatcher;
-    if (!rcc->channel) {
+    RedChannel *channel = red_channel_client_get_channel(rcc);
+    uint32_t type, id;
+
+    if (!channel) {
         return;
     }
-    dispatcher = (Dispatcher *)rcc->channel->data;
-    spice_printerr("channel type %u id %u", rcc->channel->type, rcc->channel->id);
+    g_object_get(channel, "channel-type", &type, "id", &id, NULL);
+    dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
+    spice_printerr("channel type %u id %u", type, id);
     payload.rcc = rcc;
     dispatcher_send_message(dispatcher,
                             RED_WORKER_MESSAGE_DISPLAY_MIGRATE,
@@ -135,23 +135,16 @@ static void red_qxl_display_migrate(RedChannelClient *rcc)
 }
 
 static void red_qxl_set_cursor_peer(RedChannel *channel, RedClient *client, RedsStream *stream,
-                                    int migration, int num_common_caps,
-                                    uint32_t *common_caps, int num_caps,
-                                    uint32_t *caps)
+                                    int migration,
+                                    RedChannelCapabilities *caps)
 {
     RedWorkerMessageCursorConnect payload = {0,};
-    Dispatcher *dispatcher = (Dispatcher *)channel->data;
+    Dispatcher *dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
     spice_printerr("");
     payload.client = client;
     payload.stream = stream;
     payload.migration = migration;
-    payload.num_common_caps = num_common_caps;
-    payload.common_caps = spice_malloc(sizeof(uint32_t)*num_common_caps);
-    payload.num_caps = num_caps;
-    payload.caps = spice_malloc(sizeof(uint32_t)*num_caps);
-
-    memcpy(payload.common_caps, common_caps, sizeof(uint32_t)*num_common_caps);
-    memcpy(payload.caps, caps, sizeof(uint32_t)*num_caps);
+    red_channel_capabilities_init(&payload.caps, caps);
 
     dispatcher_send_message(dispatcher,
                             RED_WORKER_MESSAGE_CURSOR_CONNECT,
@@ -162,12 +155,13 @@ static void red_qxl_disconnect_cursor_peer(RedChannelClient *rcc)
 {
     RedWorkerMessageCursorDisconnect payload;
     Dispatcher *dispatcher;
+    RedChannel *channel = red_channel_client_get_channel(rcc);
 
-    if (!rcc->channel) {
+    if (!channel) {
         return;
     }
 
-    dispatcher = (Dispatcher *)rcc->channel->data;
+    dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
     spice_printerr("");
     payload.rcc = rcc;
 
@@ -180,12 +174,15 @@ static void red_qxl_cursor_migrate(RedChannelClient *rcc)
 {
     RedWorkerMessageCursorMigrate payload;
     Dispatcher *dispatcher;
+    RedChannel *channel = red_channel_client_get_channel(rcc);
+    uint32_t type, id;
 
-    if (!rcc->channel) {
+    if (!channel) {
         return;
     }
-    dispatcher = (Dispatcher *)rcc->channel->data;
-    spice_printerr("channel type %u id %u", rcc->channel->type, rcc->channel->id);
+    g_object_get(channel, "channel-type", &type, "id", &id, NULL);
+    dispatcher = (Dispatcher *)g_object_get_data(G_OBJECT(channel), "dispatcher");
+    spice_printerr("channel type %u id %u", type, id);
     payload.rcc = rcc;
     dispatcher_send_message(dispatcher,
                             RED_WORKER_MESSAGE_CURSOR_MIGRATE,
@@ -896,11 +893,13 @@ void spice_qxl_gl_draw_async(QXLInstance *qxl,
 {
     QXLState *qxl_state;
     RedWorkerMessage message = RED_WORKER_MESSAGE_GL_DRAW_ASYNC;
-    SpiceMsgDisplayGlDraw draw = {
-        .x = x,
-        .y = y,
-        .w = w,
-        .h = h
+    RedWorkerMessageGlDraw draw = {
+        {
+            .x = x,
+            .y = y,
+            .w = w,
+            .h = h
+        },
     };
 
     spice_return_if_fail(qxl != NULL);
@@ -996,11 +995,29 @@ void red_qxl_init(RedsState *reds, QXLInstance *qxl)
     client_display_cbs.disconnect = red_qxl_disconnect_display_peer;
     client_display_cbs.migrate = red_qxl_display_migrate;
 
-    // TODO: reference and free
-    RedWorker *worker = red_worker_new(qxl, &client_cursor_cbs,
+    qxl_state->worker = red_worker_new(qxl, &client_cursor_cbs,
                                        &client_display_cbs);
 
-    red_worker_run(worker);
+    red_worker_run(qxl_state->worker);
+}
+
+void red_qxl_destroy(QXLInstance *qxl)
+{
+    spice_return_if_fail(qxl->st != NULL && qxl->st->dispatcher != NULL);
+
+    QXLState *qxl_state = qxl->st;
+
+    /* send message to close thread */
+    RedWorkerMessageClose message;
+    dispatcher_send_message(qxl_state->dispatcher,
+                            RED_WORKER_MESSAGE_CLOSE_WORKER,
+                            &message);
+    red_worker_free(qxl_state->worker);
+    g_object_unref(qxl_state->dispatcher);
+    /* this must be done after calling red_worker_free */
+    qxl->st = NULL;
+    pthread_mutex_destroy(&qxl_state->scanout_mutex);
+    free(qxl_state);
 }
 
 Dispatcher *red_qxl_get_dispatcher(QXLInstance *qxl)
