@@ -15,8 +15,8 @@
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
 
-#ifndef _H_RED_DISPATCHER
-#define _H_RED_DISPATCHER
+#ifndef RED_DISPATCHER_H_
+#define RED_DISPATCHER_H_
 
 #include "red-channel.h"
 #include "spice-qxl.h"
@@ -24,6 +24,7 @@
 typedef struct AsyncCommand AsyncCommand;
 
 void red_qxl_init(SpiceServer *reds, QXLInstance *qxl);
+void red_qxl_destroy(QXLInstance *qxl);
 
 void red_qxl_on_ic_change(QXLInstance *qxl, SpiceImageCompression ic);
 void red_qxl_on_sv_change(QXLInstance *qxl, int sv);
@@ -116,17 +117,17 @@ enum {
     RED_WORKER_MESSAGE_GL_DRAW_ASYNC,
     RED_WORKER_MESSAGE_SET_VIDEO_CODECS,
 
+    /* close worker thread */
+    RED_WORKER_MESSAGE_CLOSE_WORKER,
+
     RED_WORKER_MESSAGE_COUNT // LAST
 };
 
 typedef struct RedWorkerMessageDisplayConnect {
     RedClient * client;
     RedsStream * stream;
-    uint32_t *common_caps; // red_worker should free
-    uint32_t *caps;        // red_worker should free
+    RedChannelCapabilities caps;   // red_worker should reset
     int migration;
-    int num_common_caps;
-    int num_caps;
 } RedWorkerMessageDisplayConnect;
 
 typedef struct RedWorkerMessageDisplayDisconnect {
@@ -141,10 +142,7 @@ typedef struct RedWorkerMessageCursorConnect {
     RedClient *client;
     RedsStream *stream;
     int migration;
-    uint32_t *common_caps; // red_worker should free
-    int num_common_caps;
-    uint32_t *caps;        // red_worker should free
-    int num_caps;
+    RedChannelCapabilities caps;   // red_worker should reset
 } RedWorkerMessageCursorConnect;
 
 typedef struct RedWorkerMessageCursorDisconnect {
@@ -291,6 +289,13 @@ typedef struct RedWorkerMessageDriverUnload {
 typedef struct RedWorkerMessageGlScanout {
 } RedWorkerMessageGlScanout;
 
+typedef struct RedWorkerMessageClose {
+} RedWorkerMessageClose;
+
+typedef struct RedWorkerMessageGlDraw {
+    SpiceMsgDisplayGlDraw draw;
+} RedWorkerMessageGlDraw;
+
 enum {
     RED_DISPATCHER_PENDING_WAKEUP,
     RED_DISPATCHER_PENDING_OOM,
@@ -298,4 +303,4 @@ enum {
 
 void red_qxl_clear_pending(QXLState *qxl_state, int pending);
 
-#endif
+#endif /* RED_DISPATCHER_H_ */
