@@ -19,73 +19,105 @@
 #include <config.h>
 #endif
 
-#include <stdbool.h>
 #include <inttypes.h>
 #include <zlib.h>
 #include <pthread.h>
+#include <glib.h>
+
 #include "reds.h"
-#include "red-worker.h"
+#include "red-qxl.h"
 #include "red-common.h"
 #include "memslot.h"
 #include "red-parse-qxl.h"
-#include "red-replay-qxl.h"
-#include <glib.h>
 
 #define QXLPHYSICAL_FROM_PTR(ptr) ((QXLPHYSICAL)(intptr_t)(ptr))
 #define QXLPHYSICAL_TO_PTR(phy) ((void*)(intptr_t)(phy))
 
 typedef enum {
     REPLAY_OK = 0,
-    REPLAY_EOF,
+    REPLAY_ERROR,
 } replay_t;
 
 struct SpiceReplay {
     FILE *fd;
-    int eof;
+    gboolean error;
     int counter;
     bool created_primary;
 
     GArray *id_map; // record id -> replay id
     GArray *id_map_inv; // replay id -> record id
     GArray *id_free; // free list
+    uint8_t *primary_mem;
     int nsurfaces;
+    int end_pos;
+
+    GList *allocated;
 
     pthread_mutex_t mutex;
     pthread_cond_t cond;
 };
 
-static int replay_fread(SpiceReplay *replay, uint8_t *buf, size_t size)
+static ssize_t replay_fread(SpiceReplay *replay, uint8_t *buf, size_t size)
 {
-    if (replay->eof) {
+    if (replay->error || feof(replay->fd) ||
+        fread(buf, 1, size, replay->fd) != size) {
+        replay->error = TRUE;
         return 0;
     }
-    if (feof(replay->fd)) {
-        replay->eof = 1;
-        return 0;
-    }
-    return fread(buf, size, 1, replay->fd);
+    return size;
 }
 
 __attribute__((format(scanf, 2, 3)))
-static replay_t replay_fscanf(SpiceReplay *replay, const char *fmt, ...)
+static replay_t replay_fscanf_check(SpiceReplay *replay, const char *fmt, ...)
 {
     va_list ap;
     int ret;
 
-    if (replay->eof) {
-        return REPLAY_EOF;
+    replay->end_pos = -1;
+
+    if (replay->error) {
+        return REPLAY_ERROR;
     }
     if (feof(replay->fd)) {
-        replay->eof = 1;
-        return REPLAY_EOF;
+        replay->error = TRUE;
+        return REPLAY_ERROR;
     }
     va_start(ap, fmt);
     ret = vfscanf(replay->fd, fmt, ap);
     va_end(ap);
-    if (ret == EOF) {
-        replay->eof = 1;
+    if (ret == EOF || replay->end_pos < 0) {
+        replay->error = TRUE;
     }
-    return replay->eof ? REPLAY_EOF : REPLAY_OK;
+    return replay->error ? REPLAY_ERROR : REPLAY_OK;
+}
+#define replay_fscanf(r, fmt, ...) \
+    replay_fscanf_check(r, fmt "%n", ## __VA_ARGS__, &r->end_pos)
+
+static inline void *replay_malloc(SpiceReplay *replay, size_t size)
+{
+    void *mem = spice_malloc(size);
+    replay->allocated = g_list_prepend(replay->allocated, mem);
+    return mem;
+}
+
+static inline void *replay_malloc0(SpiceReplay *replay, size_t size)
+{
+    void *mem = replay_malloc(replay, size);
+    memset(mem, 0, size);
+    return mem;
+}
+
+static inline void replay_free(SpiceReplay *replay, void *mem)
+{
+    replay->allocated = g_list_remove(replay->allocated, mem);
+    free(mem);
+}
+
+static inline void *replay_realloc(SpiceReplay *replay, void *mem, size_t n_bytes)
+{
+    GList *elem = g_list_find(replay->allocated, mem);
+    elem->data = spice_realloc(mem, n_bytes);
+    return elem->data;
 }
 
 static uint32_t replay_id_get(SpiceReplay *replay, uint32_t id)
@@ -190,16 +222,18 @@ static replay_t read_binary(SpiceReplay *replay, const char *prefix, size_t *siz
 {
     char template[1024];
     int with_zlib = -1;
-    int zlib_size;
+    unsigned int zlib_size;
     uint8_t *zlib_buffer;
     z_stream strm;
 
-    snprintf(template, sizeof(template), "binary %%d %s %%ld:", prefix);
-    if (replay_fscanf(replay, template, &with_zlib, size) == REPLAY_EOF)
-        return REPLAY_EOF;
+    snprintf(template, sizeof(template), "binary %%d %s %%ld:%%n", prefix);
+    replay_fscanf_check(replay, template, &with_zlib, size, &replay->end_pos);
+    if (replay->error) {
+        return REPLAY_ERROR;
+    }
 
     if (*buf == NULL) {
-        *buf = spice_malloc(*size + base_size);
+        *buf = replay_malloc(replay, *size + base_size);
     }
 #if 0
     {
@@ -208,13 +242,17 @@ static replay_t read_binary(SpiceReplay *replay, const char *prefix, size_t *siz
         hexdump(*buf + base_size, *size);
     }
 #else
-    spice_return_val_if_fail(with_zlib != -1, REPLAY_EOF);
     if (with_zlib) {
         int ret;
 
-        replay_fscanf(replay, "%d:", &zlib_size);
-        zlib_buffer = spice_malloc(zlib_size);
-        replay_fread(replay, zlib_buffer, zlib_size);
+        replay_fscanf(replay, "%u:", &zlib_size);
+        if (replay->error) {
+            return REPLAY_ERROR;
+        }
+        zlib_buffer = replay_malloc(replay, zlib_size);
+        if (replay_fread(replay, zlib_buffer, zlib_size) != zlib_size) {
+            return REPLAY_ERROR;
+        }
         strm.zalloc = Z_NULL;
         strm.zfree = Z_NULL;
         strm.opaque = Z_NULL;
@@ -234,37 +272,39 @@ static replay_t read_binary(SpiceReplay *replay, const char *prefix, size_t *siz
                  * it seems it may kill the red_worker thread (so a chunk is
                  * left hanging and the rest of the message is never written).
                  * Let it pass */
-                return REPLAY_EOF;
+                return REPLAY_ERROR;
             }
             if (ret != Z_OK) {
                 spice_warn_if_reached();
             }
         }
         (void)inflateEnd(&strm);
-        free(zlib_buffer); // TODO - avoid repeat alloc/dealloc by keeping last
+        replay_free(replay, zlib_buffer); // TODO - avoid repeat alloc/dealloc by keeping last
     } else {
         replay_fread(replay, *buf + base_size, *size);
     }
 #endif
-    replay_fscanf(replay, "\n");
-    return REPLAY_OK;
+    return replay_fscanf(replay, "\n");
 }
 
-static size_t red_replay_data_chunks(SpiceReplay *replay, const char *prefix,
-                                     uint8_t **mem, size_t base_size)
+static ssize_t red_replay_data_chunks(SpiceReplay *replay, const char *prefix,
+                                      uint8_t **mem, size_t base_size)
 {
     size_t data_size;
-    int count_chunks;
+    unsigned int count_chunks;
     size_t next_data_size;
     QXLDataChunk *cur, *next;
 
-    replay_fscanf(replay, "data_chunks %d %zu\n", &count_chunks, &data_size);
+    replay_fscanf(replay, "data_chunks %u %zu\n", &count_chunks, &data_size);
+    if (replay->error) {
+        return -1;
+    }
     if (base_size == 0) {
         base_size = sizeof(QXLDataChunk);
     }
 
-    if (read_binary(replay, prefix, &next_data_size, mem, base_size) == REPLAY_EOF) {
-        return 0;
+    if (read_binary(replay, prefix, &next_data_size, mem, base_size) == REPLAY_ERROR) {
+        return -1;
     }
     cur = (QXLDataChunk*)(*mem + base_size - sizeof(QXLDataChunk));
     cur->data_size = next_data_size;
@@ -272,8 +312,8 @@ static size_t red_replay_data_chunks(SpiceReplay *replay, const char *prefix,
     cur->next_chunk = cur->prev_chunk = 0;
     while (count_chunks-- > 0) {
         if (read_binary(replay, prefix, &next_data_size, (uint8_t**)&cur->next_chunk,
-            sizeof(QXLDataChunk)) == REPLAY_EOF) {
-            return 0;
+            sizeof(QXLDataChunk)) == REPLAY_ERROR) {
+            return -1;
         }
         data_size += next_data_size;
         next = QXLPHYSICAL_TO_PTR(cur->next_chunk);
@@ -318,16 +358,20 @@ static void red_replay_rect_ptr(SpiceReplay *replay, const char *prefix, QXLRect
 {
     char template[1024];
 
-    snprintf(template, sizeof(template), "rect %s %%d %%d %%d %%d\n", prefix);
-    replay_fscanf(replay, template, &qxl->top, &qxl->left, &qxl->bottom, &qxl->right);
+    snprintf(template, sizeof(template), "rect %s %%d %%d %%d %%d\n%%n", prefix);
+    replay_fscanf_check(replay, template, &qxl->top, &qxl->left, &qxl->bottom, &qxl->right,
+                        &replay->end_pos);
 }
 
 static QXLPath *red_replay_path(SpiceReplay *replay)
 {
     QXLPath *qxl = NULL;
-    size_t data_size;
+    ssize_t data_size;
 
     data_size = red_replay_data_chunks(replay, "path", (uint8_t**)&qxl, sizeof(QXLPath));
+    if (data_size < 0) {
+        return NULL;
+    }
     qxl->data_size = data_size;
     return qxl;
 }
@@ -342,10 +386,15 @@ static void red_replay_path_free(SpiceReplay *replay, QXLPHYSICAL p)
 static QXLClipRects *red_replay_clip_rects(SpiceReplay *replay)
 {
     QXLClipRects *qxl = NULL;
-    int num_rects;
+    unsigned int num_rects;
 
-    replay_fscanf(replay, "num_rects %d\n", &num_rects);
-    red_replay_data_chunks(replay, "clip_rects", (uint8_t**)&qxl, sizeof(QXLClipRects));
+    replay_fscanf(replay, "num_rects %u\n", &num_rects);
+    if (replay->error) {
+        return NULL;
+    }
+    if (red_replay_data_chunks(replay, "clip_rects", (uint8_t**)&qxl, sizeof(QXLClipRects)) < 0) {
+        return NULL;
+    }
     qxl->num_rects = num_rects;
     return qxl;
 }
@@ -366,23 +415,30 @@ static uint8_t *red_replay_image_data_flat(SpiceReplay *replay, size_t *size)
 static QXLImage *red_replay_image(SpiceReplay *replay, uint32_t flags)
 {
     QXLImage* qxl = NULL;
-    size_t bitmap_size, size;
+    size_t bitmap_size;
+    ssize_t size;
     uint8_t qxl_flags;
     int temp;
     int has_palette;
     int has_image;
 
     replay_fscanf(replay, "image %d\n", &has_image);
+    if (replay->error) {
+        return NULL;
+    }
     if (!has_image) {
         return NULL;
     }
 
-    qxl = spice_new0(QXLImage, 1);
-    replay_fscanf(replay, "descriptor.id %"PRIu64"\n", &qxl->descriptor.id);
+    qxl = (QXLImage*)replay_malloc0(replay, sizeof(QXLImage));
+    replay_fscanf(replay, "descriptor.id %"SCNu64"\n", &qxl->descriptor.id);
     replay_fscanf(replay, "descriptor.type %d\n", &temp); qxl->descriptor.type = temp;
     replay_fscanf(replay, "descriptor.flags %d\n", &temp); qxl->descriptor.flags = temp;
     replay_fscanf(replay, "descriptor.width %d\n", &qxl->descriptor.width);
     replay_fscanf(replay, "descriptor.height %d\n", &qxl->descriptor.height);
+    if (replay->error) {
+        return NULL;
+    }
 
     switch (qxl->descriptor.type) {
     case SPICE_IMAGE_TYPE_BITMAP:
@@ -395,13 +451,16 @@ static QXLImage *red_replay_image(SpiceReplay *replay, uint32_t flags)
         replay_fscanf(replay, "has_palette %d\n", &has_palette);
         if (has_palette) {
             QXLPalette *qp;
-            int i, num_ents;
+            unsigned int i, num_ents;
 
-            replay_fscanf(replay, "qp.num_ents %d\n", &num_ents);
-            qp = spice_malloc(sizeof(QXLPalette) + num_ents * sizeof(qp->ents[0]));
+            replay_fscanf(replay, "qp.num_ents %u\n", &num_ents);
+            if (replay->error) {
+                return NULL;
+            }
+            qp = replay_malloc(replay, sizeof(QXLPalette) + num_ents * sizeof(qp->ents[0]));
             qp->num_ents = num_ents;
             qxl->bitmap.palette = QXLPHYSICAL_FROM_PTR(qp);
-            replay_fscanf(replay, "unique %"PRIu64"\n", &qp->unique);
+            replay_fscanf(replay, "unique %"SCNu64"\n", &qp->unique);
             for (i = 0; i < num_ents; i++) {
                 replay_fscanf(replay, "ents %d\n", &qp->ents[i]);
             }
@@ -422,14 +481,20 @@ static QXLImage *red_replay_image(SpiceReplay *replay, uint32_t flags)
         break;
     case SPICE_IMAGE_TYPE_SURFACE:
         replay_fscanf(replay, "surface_image.surface_id %d\n", &qxl->surface_image.surface_id);
+        if (replay->error) {
+            return NULL;
+        }
         qxl->surface_image.surface_id = replay_id_get(replay, qxl->surface_image.surface_id);
         break;
     case SPICE_IMAGE_TYPE_QUIC:
         // TODO - make this much nicer (precompute size and allocs, store them during
         // record, then reread into them. and use MPEG-4).
         replay_fscanf(replay, "quic.data_size %d\n", &qxl->quic.data_size);
-        qxl = realloc(qxl, sizeof(QXLImageDescriptor) + sizeof(QXLQUICData) +
-                      qxl->quic.data_size);
+        if (replay->error) {
+            return NULL;
+        }
+        qxl = replay_realloc(replay, qxl, sizeof(QXLImageDescriptor) + sizeof(QXLQUICData) +
+                             qxl->quic.data_size);
         size = red_replay_data_chunks(replay, "quic.data", (uint8_t**)&qxl->quic.data, 0);
         spice_assert(size == qxl->quic.data_size);
         break;
@@ -458,6 +523,7 @@ static void red_replay_image_free(SpiceReplay *replay, QXLPHYSICAL p, uint32_t f
         break;
     case SPICE_IMAGE_TYPE_QUIC:
         red_replay_data_chunks_free(replay, qxl, 0);
+        qxl = NULL;
         break;
     default:
         spice_warn_if_reached();
@@ -469,6 +535,10 @@ static void red_replay_image_free(SpiceReplay *replay, QXLPHYSICAL p, uint32_t f
 static void red_replay_brush_ptr(SpiceReplay *replay, QXLBrush *qxl, uint32_t flags)
 {
     replay_fscanf(replay, "type %d\n", &qxl->type);
+    if (replay->error) {
+        return;
+    }
+
     switch (qxl->type) {
     case SPICE_BRUSH_TYPE_SOLID:
         replay_fscanf(replay, "u.color %d\n", &qxl->u.color);
@@ -633,6 +703,9 @@ static void red_replay_stroke_ptr(SpiceReplay *replay, QXLStroke *qxl, uint32_t
 
     qxl->path = QXLPHYSICAL_FROM_PTR(red_replay_path(replay));
     replay_fscanf(replay, "attr.flags %d\n", &temp); qxl->attr.flags = temp;
+    if (replay->error) {
+        return;
+    }
     if (qxl->attr.flags & SPICE_LINE_FLAGS_STYLED) {
         size_t size;
 
@@ -659,13 +732,16 @@ static QXLString *red_replay_string(SpiceReplay *replay)
     uint32_t data_size;
     uint16_t length;
     uint16_t flags;
-    size_t chunk_size;
+    ssize_t chunk_size;
     QXLString *qxl = NULL;
 
     replay_fscanf(replay, "data_size %d\n", &data_size);
     replay_fscanf(replay, "length %d\n", &temp); length = temp;
     replay_fscanf(replay, "flags %d\n", &temp); flags = temp;
     chunk_size = red_replay_data_chunks(replay, "string", (uint8_t**)&qxl, sizeof(QXLString));
+    if (chunk_size <  0) {
+        return NULL;
+    }
     qxl->data_size = data_size;
     qxl->length = length;
     qxl->flags = flags;
@@ -730,6 +806,9 @@ static void red_replay_invers_free(SpiceReplay *replay, QXLInvers *qxl, uint32_t
 static void red_replay_clip_ptr(SpiceReplay *replay, QXLClip *qxl)
 {
     replay_fscanf(replay, "type %d\n", &qxl->type);
+    if (replay->error) {
+        return;
+    }
     switch (qxl->type) {
     case SPICE_CLIP_TYPE_RECTS:
         qxl->data = QXLPHYSICAL_FROM_PTR(red_replay_clip_rects(replay));
@@ -759,7 +838,7 @@ static uint8_t *red_replay_transform(SpiceReplay *replay)
 
 static void red_replay_composite_ptr(SpiceReplay *replay, QXLComposite *qxl, uint32_t flags)
 {
-    int enabled;
+    int enabled = 0;
 
     replay_fscanf(replay, "flags %d\n", &qxl->flags);
     qxl->src = QXLPHYSICAL_FROM_PTR(red_replay_image(replay, flags));
@@ -788,7 +867,7 @@ static void red_replay_composite_free(SpiceReplay *replay, QXLComposite *qxl, ui
 
 static QXLDrawable *red_replay_native_drawable(SpiceReplay *replay, uint32_t flags)
 {
-    QXLDrawable *qxl = spice_malloc0(sizeof(QXLDrawable)); // TODO - this is too large usually
+    QXLDrawable *qxl = replay_malloc0(replay, sizeof(QXLDrawable)); // TODO - this is too large usually
     int i;
     int temp;
 
@@ -799,15 +878,24 @@ static QXLDrawable *red_replay_native_drawable(SpiceReplay *replay, uint32_t fla
     replay_fscanf(replay, "self_bitmap %d\n", &temp); qxl->self_bitmap = temp;
     red_replay_rect_ptr(replay, "self_bitmap_area", &qxl->self_bitmap_area);
     replay_fscanf(replay, "surface_id %d\n", &qxl->surface_id);
+    if (replay->error) {
+        return NULL;
+    }
     qxl->surface_id = replay_id_get(replay, qxl->surface_id);
 
     for (i = 0; i < 3; i++) {
         replay_fscanf(replay, "surfaces_dest %d\n", &qxl->surfaces_dest[i]);
+        if (replay->error) {
+            return NULL;
+        }
         qxl->surfaces_dest[i] = replay_id_get(replay, qxl->surfaces_dest[i]);
         red_replay_rect_ptr(replay, "surfaces_rects", &qxl->surfaces_rects[i]);
     }
 
     replay_fscanf(replay, "type %d\n", &temp); qxl->type = temp;
+    if (replay->error) {
+        return NULL;
+    }
     switch (qxl->type) {
     case QXL_DRAW_ALPHA_BLEND:
         red_replay_alpha_blend_ptr(replay, &qxl->u.alpha_blend, flags);
@@ -919,7 +1007,7 @@ static void red_replay_native_drawable_free(SpiceReplay *replay, QXLDrawable *qx
 static QXLCompatDrawable *red_replay_compat_drawable(SpiceReplay *replay, uint32_t flags)
 {
     int temp;
-    QXLCompatDrawable *qxl = spice_malloc0(sizeof(QXLCompatDrawable)); // TODO - too large usually
+    QXLCompatDrawable *qxl = replay_malloc0(replay, sizeof(QXLCompatDrawable)); // TODO - too large usually
 
     red_replay_rect_ptr(replay, "bbox", &qxl->bbox);
     red_replay_clip_ptr(replay, &qxl->clip);
@@ -930,6 +1018,9 @@ static QXLCompatDrawable *red_replay_compat_drawable(SpiceReplay *replay, uint32
     red_replay_rect_ptr(replay, "bitmap_area", &qxl->bitmap_area);
 
     replay_fscanf(replay, "type %d\n", &temp); qxl->type = temp;
+    if (replay->error) {
+        return NULL;
+    }
     switch (qxl->type) {
     case QXL_DRAW_ALPHA_BLEND:
         red_replay_alpha_blend_ptr_compat(replay, &qxl->u.alpha_blend, flags);
@@ -973,7 +1064,7 @@ static QXLCompatDrawable *red_replay_compat_drawable(SpiceReplay *replay, uint32
         red_replay_whiteness_ptr(replay, &qxl->u.whiteness, flags);
         break;
     default:
-        spice_error("%s: unknown type %d", __FUNCTION__, qxl->type);
+        spice_error("unknown type %d", qxl->type);
         break;
     };
     return qxl;
@@ -981,10 +1072,10 @@ static QXLCompatDrawable *red_replay_compat_drawable(SpiceReplay *replay, uint32
 
 static QXLPHYSICAL red_replay_drawable(SpiceReplay *replay, uint32_t flags)
 {
-    if (replay->eof) {
+    replay_fscanf(replay, "drawable\n");
+    if (replay->error) {
         return 0;
     }
-    replay_fscanf(replay, "drawable\n");
     if (flags & QXL_COMMAND_FLAG_COMPAT) {
         return QXLPHYSICAL_FROM_PTR(red_replay_compat_drawable(replay, flags));
     } else {
@@ -994,12 +1085,15 @@ static QXLPHYSICAL red_replay_drawable(SpiceReplay *replay, uint32_t flags)
 
 static QXLUpdateCmd *red_replay_update_cmd(SpiceReplay *replay)
 {
-    QXLUpdateCmd *qxl = spice_malloc0(sizeof(QXLUpdateCmd));
+    QXLUpdateCmd *qxl = replay_malloc0(replay, sizeof(QXLUpdateCmd));
 
     replay_fscanf(replay, "update\n");
     red_replay_rect_ptr(replay, "area", &qxl->area);
     replay_fscanf(replay, "update_id %d\n", &qxl->update_id);
     replay_fscanf(replay, "surface_id %d\n", &qxl->surface_id);
+    if (replay->error) {
+        return NULL;
+    }
     qxl->surface_id = replay_id_get(replay, qxl->surface_id);
 
     return qxl;
@@ -1007,7 +1101,7 @@ static QXLUpdateCmd *red_replay_update_cmd(SpiceReplay *replay)
 
 static QXLMessage *red_replay_message(SpiceReplay *replay)
 {
-    QXLMessage *qxl;
+    QXLMessage *qxl = NULL;
     size_t size;
 
     read_binary(replay, "message", &size, (uint8_t**)&qxl, sizeof(QXLMessage));
@@ -1019,12 +1113,15 @@ static QXLSurfaceCmd *red_replay_surface_cmd(SpiceReplay *replay)
     size_t size;
     size_t read_size;
     int temp;
-    QXLSurfaceCmd *qxl = spice_malloc0(sizeof(QXLSurfaceCmd));
+    QXLSurfaceCmd *qxl = replay_malloc0(replay, sizeof(QXLSurfaceCmd));
 
     replay_fscanf(replay, "surface_cmd\n");
     replay_fscanf(replay, "surface_id %d\n", &qxl->surface_id);
     replay_fscanf(replay, "type %d\n", &temp); qxl->type = temp;
     replay_fscanf(replay, "flags %d\n", &qxl->flags);
+    if (replay->error) {
+        return NULL;
+    }
 
     switch (qxl->type) {
     case QXL_SURFACE_CMD_CREATE:
@@ -1032,6 +1129,9 @@ static QXLSurfaceCmd *red_replay_surface_cmd(SpiceReplay *replay)
         replay_fscanf(replay, "u.surface_create.width %d\n", &qxl->u.surface_create.width);
         replay_fscanf(replay, "u.surface_create.height %d\n", &qxl->u.surface_create.height);
         replay_fscanf(replay, "u.surface_create.stride %d\n", &qxl->u.surface_create.stride);
+        if (replay->error) {
+            return NULL;
+        }
         size = qxl->u.surface_create.height * abs(qxl->u.surface_create.stride);
         if ((qxl->flags & QXL_SURF_FLAG_KEEP_DATA) != 0) {
             read_binary(replay, "data", &read_size, (uint8_t**)&qxl->u.surface_create.data, 0);
@@ -1039,7 +1139,7 @@ static QXLSurfaceCmd *red_replay_surface_cmd(SpiceReplay *replay)
                 spice_printerr("mismatch %zu != %zu", size, read_size);
             }
         } else {
-            qxl->u.surface_create.data = QXLPHYSICAL_FROM_PTR(spice_malloc(size));
+            qxl->u.surface_create.data = QXLPHYSICAL_FROM_PTR(replay_malloc(replay, size));
         }
         qxl->surface_id = replay_id_new(replay, qxl->surface_id);
         break;
@@ -1064,6 +1164,7 @@ static QXLCursor *red_replay_cursor(SpiceReplay *replay)
 {
     int temp;
     QXLCursor cursor, *qxl = NULL;
+    ssize_t data_size;
 
     replay_fscanf(replay, "header.unique %"SCNu64"\n", &cursor.header.unique);
     replay_fscanf(replay, "header.type %d\n", &temp);
@@ -1078,20 +1179,28 @@ static QXLCursor *red_replay_cursor(SpiceReplay *replay)
     cursor.header.hot_spot_y = temp;
 
     replay_fscanf(replay, "data_size %d\n", &temp);
-    cursor.data_size = temp;
-    cursor.data_size = red_replay_data_chunks(replay, "cursor", (uint8_t**)&qxl, sizeof(QXLCursor));
+    if (replay->error) {
+        return NULL;
+    }
+    data_size = red_replay_data_chunks(replay, "cursor", (uint8_t**)&qxl, sizeof(QXLCursor));
+    if (data_size < 0) {
+        return NULL;
+    }
     qxl->header = cursor.header;
-    qxl->data_size = cursor.data_size;
+    qxl->data_size = data_size;
     return qxl;
 }
 
 static QXLCursorCmd *red_replay_cursor_cmd(SpiceReplay *replay)
 {
     int temp;
-    QXLCursorCmd *qxl = spice_new0(QXLCursorCmd, 1);
+    QXLCursorCmd *qxl = replay_malloc0(replay, sizeof(QXLCursorCmd));
 
     replay_fscanf(replay, "cursor_cmd\n");
     replay_fscanf(replay, "type %d\n", &temp);
+    if (replay->error) {
+        return NULL;
+    }
     qxl->type = temp;
     switch (qxl->type) {
     case QXL_CURSOR_SET:
@@ -1141,8 +1250,14 @@ static void replay_handle_create_primary(QXLWorker *worker, SpiceReplay *replay)
         &surface.stride, &surface.format);
     replay_fscanf(replay, "%d %d %d %d\n", &surface.position, &surface.mouse_mode,
         &surface.flags, &surface.type);
+    if (replay->error) {
+        return;
+    }
     read_binary(replay, "data", &size, &mem, 0);
     surface.group_id = 0;
+    free(replay->primary_mem);
+    replay->allocated = g_list_remove(replay->allocated, mem);
+    replay->primary_mem = mem;
     surface.mem = QXLPHYSICAL_FROM_PTR(mem);
     worker->create_primary_surface(worker, 0, &surface);
 }
@@ -1158,6 +1273,8 @@ static void replay_handle_dev_input(QXLWorker *worker, SpiceReplay *replay,
     case RED_WORKER_MESSAGE_DESTROY_PRIMARY_SURFACE:
         replay->created_primary = FALSE;
         worker->destroy_primary_surface(worker, 0);
+        free(replay->primary_mem);
+        replay->primary_mem = NULL;
         break;
     case RED_WORKER_MESSAGE_DESTROY_SURFACES:
         replay->created_primary = FALSE;
@@ -1178,33 +1295,33 @@ static void replay_handle_dev_input(QXLWorker *worker, SpiceReplay *replay,
 
 /*
  * NOTE: This reads from a saved file and performs all io actions, calling the
- * dispatcher, until it sees a command, at which point it returns it via the
- * last parameter [ext_cmd]. Hence you cannot call this from the worker thread
+ * dispatcher, until it sees a command, at which point it returns it.
+ * Hence you cannot call this from the worker thread
  * since it will block reading from the dispatcher pipe.
  */
 SPICE_GNUC_VISIBLE QXLCommandExt* spice_replay_next_cmd(SpiceReplay *replay,
                                                          QXLWorker *worker)
 {
-    QXLCommandExt* cmd;
+    QXLCommandExt* cmd = NULL;
     uint64_t timestamp;
     int type;
     int what = -1;
     int counter;
 
     while (what != 0) {
-        replay_fscanf(replay, "event %d %d %d %"PRIu64"\n", &counter,
+        replay_fscanf(replay, "event %d %d %d %"SCNu64"\n", &counter,
                             &what, &type, &timestamp);
-        if (replay->eof) {
-            return NULL;
+        if (replay->error) {
+            goto error;
         }
         if (what == 1) {
             replay_handle_dev_input(worker, replay, type);
         }
     }
-    cmd = g_new(QXLCommandExt, 1);
+    cmd = replay_malloc0(replay, sizeof(QXLCommandExt));
     cmd->cmd.type = type;
     cmd->group_id = 0;
-    spice_debug("command %"PRIu64", %d\r", timestamp, cmd->cmd.type);
+    spice_debug("command %"SCNu64", %d\r", timestamp, cmd->cmd.type);
     switch (cmd->cmd.type) {
     case QXL_CMD_DRAW:
         cmd->flags = 0;
@@ -1224,6 +1341,10 @@ SPICE_GNUC_VISIBLE QXLCommandExt* spice_replay_next_cmd(SpiceReplay *replay,
         break;
     }
 
+    if (replay->error) {
+        goto error;
+    }
+
     QXLReleaseInfo *info;
     switch (cmd->cmd.type) {
     case QXL_CMD_DRAW:
@@ -1234,9 +1355,25 @@ SPICE_GNUC_VISIBLE QXLCommandExt* spice_replay_next_cmd(SpiceReplay *replay,
         info->id = (uintptr_t)cmd;
     }
 
+    /* all buffer allocated will be used by the caller but
+     * free the list of buffer allocated to avoid to free on next calls */
+    if (replay->allocated) {
+        g_list_free(replay->allocated);
+        replay->allocated = NULL;
+    }
+
     replay->counter++;
 
     return cmd;
+
+error:
+    /* if there were some allocation during the read free all
+     * buffers allocated to avoid leaks */
+    if (replay->allocated) {
+        g_list_free_full(replay->allocated, free);
+        replay->allocated = NULL;
+    }
+    return NULL;
 }
 
 SPICE_GNUC_VISIBLE void spice_replay_free_cmd(SpiceReplay *replay, QXLCommandExt *cmd)
@@ -1271,7 +1408,7 @@ SPICE_GNUC_VISIBLE void spice_replay_free_cmd(SpiceReplay *replay, QXLCommandExt
         break;
     }
 
-    g_free(cmd);
+    free(cmd);
 }
 
 /* caller is incharge of closing the replay when done and releasing the SpiceReplay
@@ -1285,7 +1422,7 @@ SpiceReplay *spice_replay_new(FILE *file, int nsurfaces)
     spice_return_val_if_fail(file != NULL, NULL);
 
     if (fscanf(file, "SPICE_REPLAY %u\n", &version) == 1) {
-        if (version > 1) {
+        if (version != 1) {
             spice_warning("Replay file version unsupported");
             return NULL;
         }
@@ -1296,7 +1433,7 @@ SpiceReplay *spice_replay_new(FILE *file, int nsurfaces)
 
     replay = spice_malloc0(sizeof(SpiceReplay));
 
-    replay->eof = 0;
+    replay->error = FALSE;
     replay->fd = file;
     replay->created_primary = FALSE;
     pthread_mutex_init(&replay->mutex, NULL);
@@ -1305,6 +1442,7 @@ SpiceReplay *spice_replay_new(FILE *file, int nsurfaces)
     replay->id_map_inv = g_array_new(FALSE, FALSE, sizeof(uint32_t));
     replay->id_free = g_array_new(FALSE, FALSE, sizeof(uint32_t));
     replay->nsurfaces = nsurfaces;
+    replay->allocated = NULL;
 
     /* reserve id 0 */
     replay_id_new(replay, 0);
@@ -1316,11 +1454,13 @@ SPICE_GNUC_VISIBLE void spice_replay_free(SpiceReplay *replay)
 {
     spice_return_if_fail(replay != NULL);
 
+    g_list_free_full(replay->allocated, free);
     pthread_mutex_destroy(&replay->mutex);
     pthread_cond_destroy(&replay->cond);
     g_array_free(replay->id_map, TRUE);
     g_array_free(replay->id_map_inv, TRUE);
     g_array_free(replay->id_free, TRUE);
+    free(replay->primary_mem);
     fclose(replay->fd);
     free(replay);
 }
