@@ -19,20 +19,14 @@
 #include <config.h>
 #endif
 
-#define SPICE_LOG_DOMAIN "SpiceWorker"
-
 #include <stdio.h>
 #include <stdarg.h>
-#include <fcntl.h>
-#include <sys/socket.h>
-#include <netinet/in.h>
 #include <stdlib.h>
 #include <errno.h>
 #include <string.h>
 #include <unistd.h>
 #include <poll.h>
 #include <pthread.h>
-#include <netinet/tcp.h>
 #include <openssl/ssl.h>
 #include <inttypes.h>
 #include <glib.h>
@@ -80,47 +74,20 @@ struct RedWorker {
     spice_wan_compression_t zlib_glz_state;
 
     uint32_t process_display_generation;
-#ifdef RED_STATISTICS
-    StatNodeRef stat;
-    uint64_t *wakeup_counter;
-    uint64_t *command_counter;
-#endif
+    RedStatNode stat;
+    RedStatCounter wakeup_counter;
+    RedStatCounter command_counter;
 
     int driver_cap_monitors_config;
 
     RedRecord *record;
+    GMainLoop *loop;
 };
 
-static RedsState* red_worker_get_server(RedWorker *worker);
-
 static int display_is_connected(RedWorker *worker)
 {
-    return (worker->display_channel && red_channel_is_connected(
-        &worker->display_channel->common.base));
-}
-
-static uint8_t *common_alloc_recv_buf(RedChannelClient *rcc, uint16_t type, uint32_t size)
-{
-    CommonGraphicsChannel *common = SPICE_CONTAINEROF(rcc->channel, CommonGraphicsChannel, base);
-
-    /* SPICE_MSGC_MIGRATE_DATA is the only client message whose size is dynamic */
-    if (type == SPICE_MSGC_MIGRATE_DATA) {
-        return spice_malloc(size);
-    }
-
-    if (size > CHANNEL_RECEIVE_BUF_SIZE) {
-        spice_critical("unexpected message size %u (max is %d)", size, CHANNEL_RECEIVE_BUF_SIZE);
-        return NULL;
-    }
-    return common->recv_buf;
-}
-
-static void common_release_recv_buf(RedChannelClient *rcc, uint16_t type, uint32_t size,
-                                    uint8_t* msg)
-{
-    if (type == SPICE_MSGC_MIGRATE_DATA) {
-        free(msg);
-    }
+    return worker->display_channel &&
+        red_channel_is_connected(RED_CHANNEL(worker->display_channel));
 }
 
 void red_drawable_unref(RedDrawable *red_drawable)
@@ -133,6 +100,19 @@ void red_drawable_unref(RedDrawable *red_drawable)
     free(red_drawable);
 }
 
+static gboolean red_process_cursor_cmd(RedWorker *worker, const QXLCommandExt *ext)
+{
+    RedCursorCmd *cursor_cmd;
+
+    cursor_cmd = spice_new0(RedCursorCmd, 1);
+    if (!red_get_cursor_cmd(&worker->mem_slots, ext->group_id, cursor_cmd, ext->cmd.data)) {
+        free(cursor_cmd);
+        return FALSE;
+    }
+    cursor_channel_process_cmd(worker->cursor_channel, cursor_cmd);
+    return TRUE;
+}
+
 static int red_process_cursor(RedWorker *worker, int *ring_is_empty)
 {
     QXLCommandExt ext_cmd;
@@ -157,24 +137,15 @@ static int red_process_cursor(RedWorker *worker, int *ring_is_empty)
             return n;
         }
 
-        if (worker->record)
-            red_record_qxl_command(worker->record, &worker->mem_slots, ext_cmd,
-                                   spice_get_monotonic_time_ns());
+        if (worker->record) {
+            red_record_qxl_command(worker->record, &worker->mem_slots, ext_cmd);
+        }
 
         worker->cursor_poll_tries = 0;
         switch (ext_cmd.cmd.type) {
-        case QXL_CMD_CURSOR: {
-            RedCursorCmd *cursor = spice_new0(RedCursorCmd, 1);
-
-            if (red_get_cursor_cmd(&worker->mem_slots, ext_cmd.group_id,
-                                    cursor, ext_cmd.cmd.data)) {
-                free(cursor);
-                break;
-            }
-
-            cursor_channel_process_cmd(worker->cursor_channel, cursor);
+        case QXL_CMD_CURSOR:
+            red_process_cursor_cmd(worker, &ext_cmd);
             break;
-        }
         default:
             spice_warning("bad command type");
         }
@@ -194,6 +165,20 @@ static RedDrawable *red_drawable_new(QXLInstance *qxl)
     return red;
 }
 
+static gboolean red_process_surface_cmd(RedWorker *worker, QXLCommandExt *ext, gboolean loadvm)
+{
+    RedSurfaceCmd surface_cmd;
+
+    if (!red_get_surface_cmd(&worker->mem_slots, ext->group_id, &surface_cmd, ext->cmd.data)) {
+        return FALSE;
+    }
+    display_channel_process_surface_cmd(worker->display_channel, &surface_cmd, loadvm);
+    // display_channel_process_surface_cmd() takes ownership of 'release_info_ext',
+    // we don't need to release it ourselves
+    red_put_surface_cmd(&surface_cmd);
+    return TRUE;
+}
+
 static int red_process_display(RedWorker *worker, int *ring_is_empty)
 {
     QXLCommandExt ext_cmd;
@@ -220,17 +205,17 @@ static int red_process_display(RedWorker *worker, int *ring_is_empty)
             return n;
         }
 
-        if (worker->record)
-            red_record_qxl_command(worker->record, &worker->mem_slots, ext_cmd,
-                                   spice_get_monotonic_time_ns());
+        if (worker->record) {
+            red_record_qxl_command(worker->record, &worker->mem_slots, ext_cmd);
+        }
 
-        stat_inc_counter(reds, worker->command_counter, 1);
+        stat_inc_counter(worker->command_counter, 1);
         worker->display_poll_tries = 0;
         switch (ext_cmd.cmd.type) {
         case QXL_CMD_DRAW: {
             RedDrawable *red_drawable = red_drawable_new(worker->qxl); // returns with 1 ref
 
-            if (!red_get_drawable(&worker->mem_slots, ext_cmd.group_id,
+            if (red_get_drawable(&worker->mem_slots, ext_cmd.group_id,
                                  red_drawable, ext_cmd.cmd.data, ext_cmd.flags)) {
                 display_channel_process_draw(worker->display_channel, red_drawable,
                                              worker->process_display_generation);
@@ -242,11 +227,11 @@ static int red_process_display(RedWorker *worker, int *ring_is_empty)
         case QXL_CMD_UPDATE: {
             RedUpdateCmd update;
 
-            if (red_get_update_cmd(&worker->mem_slots, ext_cmd.group_id,
-                                   &update, ext_cmd.cmd.data)) {
+            if (!red_get_update_cmd(&worker->mem_slots, ext_cmd.group_id,
+                                    &update, ext_cmd.cmd.data)) {
                 break;
             }
-            if (!validate_surface(worker->display_channel, update.surface_id)) {
+            if (!display_channel_validate_surface(worker->display_channel, update.surface_id)) {
                 spice_warning("Invalid surface in QXL_CMD_UPDATE");
             } else {
                 display_channel_draw(worker->display_channel, &update.area, update.surface_id);
@@ -259,35 +244,26 @@ static int red_process_display(RedWorker *worker, int *ring_is_empty)
         case QXL_CMD_MESSAGE: {
             RedMessage message;
 
-            if (red_get_message(&worker->mem_slots, ext_cmd.group_id,
-                                &message, ext_cmd.cmd.data)) {
+            if (!red_get_message(&worker->mem_slots, ext_cmd.group_id,
+                                 &message, ext_cmd.cmd.data)) {
                 break;
             }
 #ifdef DEBUG
-            /* alert: accessing message.data is insecure */
-            spice_warning("MESSAGE: %s", message.data);
+            spice_warning("MESSAGE: %.*s", message.len, message.data);
 #endif
             red_qxl_release_resource(worker->qxl, message.release_info_ext);
             red_put_message(&message);
             break;
         }
-        case QXL_CMD_SURFACE: {
-            RedSurfaceCmd surface;
-
-            if (red_get_surface_cmd(&worker->mem_slots, ext_cmd.group_id,
-                                    &surface, ext_cmd.cmd.data)) {
-                break;
-            }
-            display_channel_process_surface_cmd(worker->display_channel, &surface, FALSE);
-            // do not release resource as is released inside display_channel_process_surface_cmd
-            red_put_surface_cmd(&surface);
+        case QXL_CMD_SURFACE:
+            red_process_surface_cmd(worker, &ext_cmd, FALSE);
             break;
-        }
+
         default:
             spice_error("bad command type");
         }
         n++;
-        if (red_channel_all_blocked(&worker->display_channel->common.base)
+        if (red_channel_all_blocked(RED_CHANNEL(worker->display_channel))
             || spice_get_monotonic_time_ns() - start > NSEC_PER_SEC / 100) {
             worker->event_timeout = 0;
             return n;
@@ -400,111 +376,11 @@ static void flush_all_qxl_commands(RedWorker *worker)
     flush_cursor_commands(worker);
 }
 
-static int common_channel_config_socket(RedChannelClient *rcc)
-{
-    RedClient *client = red_channel_client_get_client(rcc);
-    MainChannelClient *mcc = red_client_get_main(client);
-    RedsStream *stream = red_channel_client_get_stream(rcc);
-    CommonGraphicsChannelClient *ccc = COMMON_GRAPHICS_CHANNEL_CLIENT(rcc);
-    int flags;
-    int delay_val;
-
-    if ((flags = fcntl(stream->socket, F_GETFL)) == -1) {
-        spice_warning("accept failed, %s", strerror(errno));
-        return FALSE;
-    }
-
-    if (fcntl(stream->socket, F_SETFL, flags | O_NONBLOCK) == -1) {
-        spice_warning("accept failed, %s", strerror(errno));
-        return FALSE;
-    }
-
-    // TODO - this should be dynamic, not one time at channel creation
-    ccc->is_low_bandwidth = main_channel_client_is_low_bandwidth(mcc);
-    delay_val = ccc->is_low_bandwidth ? 0 : 1;
-    /* FIXME: Using Nagle's Algorithm can lead to apparent delays, depending
-     * on the delayed ack timeout on the other side.
-     * Instead of using Nagle's, we need to implement message buffering on
-     * the application level.
-     * see: http://www.stuartcheshire.org/papers/NagleDelayedAck/
-     */
-    if (setsockopt(stream->socket, IPPROTO_TCP, TCP_NODELAY, &delay_val,
-                   sizeof(delay_val)) == -1) {
-        if (errno != ENOTSUP) {
-            spice_warning("setsockopt failed, %s", strerror(errno));
-        }
-    }
-    return TRUE;
-}
-
-CommonGraphicsChannelClient *common_graphics_channel_new_client(CommonGraphicsChannel *common,
-                                                                int size,
-                                                                RedClient *client,
-                                                                RedsStream *stream,
-                                                                int mig_target,
-                                                                int monitor_latency,
-                                                                uint32_t *common_caps,
-                                                                int num_common_caps,
-                                                                uint32_t *caps,
-                                                                int num_caps)
-{
-    RedChannelClient *rcc =
-        red_channel_client_create(size, &common->base, client, stream, monitor_latency,
-                                  num_common_caps, common_caps, num_caps, caps);
-    if (!rcc) {
-        return NULL;
-    }
-    CommonGraphicsChannelClient *common_cc = (CommonGraphicsChannelClient*)rcc;
-    common->during_target_migrate = mig_target;
-
-    // TODO: move wide/narrow ack setting to red_channel.
-    red_channel_client_ack_set_client_window(rcc,
-        common_cc->is_low_bandwidth ?
-        WIDE_CLIENT_ACK_WINDOW : NARROW_CLIENT_ACK_WINDOW);
-    return common_cc;
-}
-
-
-CommonGraphicsChannel *red_worker_new_channel(RedWorker *worker, int size,
-                                              const char *name,
-                                              uint32_t channel_type, int migration_flags,
-                                              ChannelCbs *channel_cbs,
-                                              channel_handle_parsed_proc handle_parsed)
-{
-    RedChannel *channel = NULL;
-    CommonGraphicsChannel *common;
-
-    spice_return_val_if_fail(worker, NULL);
-    spice_return_val_if_fail(channel_cbs, NULL);
-    spice_return_val_if_fail(!channel_cbs->config_socket, NULL);
-    spice_return_val_if_fail(!channel_cbs->alloc_recv_buf, NULL);
-    spice_return_val_if_fail(!channel_cbs->release_recv_buf, NULL);
-
-    channel_cbs->config_socket = common_channel_config_socket;
-    channel_cbs->alloc_recv_buf = common_alloc_recv_buf;
-    channel_cbs->release_recv_buf = common_release_recv_buf;
-
-    channel = red_channel_create_parser(size, red_worker_get_server(worker),
-                                        &worker->core, channel_type,
-                                        worker->qxl->id, TRUE /* handle_acks */,
-                                        spice_get_client_channel_parser(channel_type, NULL),
-                                        handle_parsed,
-                                        channel_cbs,
-                                        migration_flags);
-    spice_return_val_if_fail(channel, NULL);
-    red_channel_set_stat_node(channel, stat_add_node(red_worker_get_server(worker),
-                                                     worker->stat, name, TRUE));
-
-    common = (CommonGraphicsChannel *)channel;
-    common->qxl = worker->qxl;
-    return common;
-}
-
 static void guest_set_client_capabilities(RedWorker *worker)
 {
     int i;
     RedChannelClient *rcc;
-    GList *link, *next;
+    GListIter iter;
     uint8_t caps[SPICE_CAPABILITIES_SIZE] = { 0 };
     int caps_available[] = {
         SPICE_DISPLAY_CAP_SIZED_STREAM,
@@ -530,15 +406,15 @@ static void guest_set_client_capabilities(RedWorker *worker)
         return;
     }
     if ((worker->display_channel == NULL) ||
-        (RED_CHANNEL(worker->display_channel)->clients == NULL)) {
+        (red_channel_get_n_clients(RED_CHANNEL(worker->display_channel)) == 0)) {
         red_qxl_set_client_capabilities(worker->qxl, FALSE, caps);
     } else {
         // Take least common denominator
-        for (i = 0 ; i < sizeof(caps_available) / sizeof(caps_available[0]); ++i) {
+        for (i = 0 ; i < SPICE_N_ELEMENTS(caps_available); ++i) {
             SET_CAP(caps, caps_available[i]);
         }
-        FOREACH_CLIENT(worker->display_channel, link, next, rcc) {
-            for (i = 0 ; i < sizeof(caps_available) / sizeof(caps_available[0]); ++i) {
+        FOREACH_CLIENT(worker->display_channel, iter, rcc) {
+            for (i = 0 ; i < SPICE_N_ELEMENTS(caps_available); ++i) {
                 if (!red_channel_client_test_remote_cap(rcc, caps_available[i]))
                     CLEAR_CAP(caps, caps_available[i]);
             }
@@ -610,43 +486,16 @@ static void handle_dev_destroy_surfaces(void *opaque, void *payload)
     cursor_channel_reset(worker->cursor_channel);
 }
 
-static void display_update_monitors_config(DisplayChannel *display,
-                                           QXLMonitorsConfig *config,
-                                           uint16_t count, uint16_t max_allowed)
-{
-
-    if (display->monitors_config)
-        monitors_config_unref(display->monitors_config);
-
-    display->monitors_config =
-        monitors_config_new(config->heads, count, max_allowed);
-}
-
 static void red_worker_push_monitors_config(RedWorker *worker)
 {
     DisplayChannelClient *dcc;
-    GList *item, *next;
+    GListIter iter;
 
-    FOREACH_CLIENT(worker->display_channel, item, next, dcc) {
+    FOREACH_DCC(worker->display_channel, iter, dcc) {
         dcc_push_monitors_config(dcc);
     }
 }
 
-static void set_monitors_config_to_primary(DisplayChannel *display)
-{
-    DrawContext *context = &display->surfaces[0].context;
-    QXLHead head = { 0, };
-
-    spice_return_if_fail(display->surfaces[0].context.canvas);
-
-    if (display->monitors_config)
-        monitors_config_unref(display->monitors_config);
-
-    head.width = context->width;
-    head.height = context->height;
-    display->monitors_config = monitors_config_new(&head, 1, 1);
-}
-
 static void dev_create_primary_surface(RedWorker *worker, uint32_t surface_id,
                                        QXLDevSurfaceCreate surface)
 {
@@ -654,7 +503,7 @@ static void dev_create_primary_surface(RedWorker *worker, uint32_t surface_id,
     uint8_t *line_0;
     int error;
 
-    spice_debug(NULL);
+    spice_debug("trace");
     spice_warn_if_fail(surface_id == 0);
     spice_warn_if_fail(surface.height != 0);
 
@@ -684,20 +533,22 @@ static void dev_create_primary_surface(RedWorker *worker, uint32_t surface_id,
 
     display_channel_create_surface(display, 0, surface.width, surface.height, surface.stride, surface.format,
                                    line_0, surface.flags & QXL_SURF_FLAG_KEEP_DATA, TRUE);
-    set_monitors_config_to_primary(display);
+    display_channel_set_monitors_config_to_primary(display);
 
-    if (display_is_connected(worker) && !worker->display_channel->common.during_target_migrate) {
+    CommonGraphicsChannel *common = COMMON_GRAPHICS_CHANNEL(worker->display_channel);
+    if (display_is_connected(worker) &&
+        !common_graphics_channel_get_during_target_migrate(common)) {
         /* guest created primary, so it will (hopefully) send a monitors_config
          * now, don't send our own temporary one */
         if (!worker->driver_cap_monitors_config) {
             red_worker_push_monitors_config(worker);
         }
-        red_pipes_add_verb(&worker->display_channel->common.base,
-                           SPICE_MSG_DISPLAY_MARK);
-        red_channel_push(&worker->display_channel->common.base);
+        red_channel_pipes_add_empty_msg(RED_CHANNEL(worker->display_channel),
+                                        SPICE_MSG_DISPLAY_MARK);
+        red_channel_push(RED_CHANNEL(worker->display_channel));
     }
 
-    cursor_channel_init(worker->cursor_channel);
+    cursor_channel_do_init(worker->cursor_channel);
 }
 
 static void handle_dev_create_primary_surface(void *opaque, void *payload)
@@ -712,22 +563,20 @@ static void destroy_primary_surface(RedWorker *worker, uint32_t surface_id)
 {
     DisplayChannel *display = worker->display_channel;
 
-    if (!validate_surface(display, surface_id))
-        return;
-    spice_warn_if_fail(surface_id == 0);
-
-    spice_debug(NULL);
-    if (!display->surfaces[surface_id].context.canvas) {
+    if (!display_channel_validate_surface(display, surface_id)) {
         spice_warning("double destroy of primary surface");
         return;
     }
+    spice_warn_if_fail(surface_id == 0);
 
     flush_all_qxl_commands(worker);
     display_channel_destroy_surface_wait(display, 0);
     display_channel_surface_unref(display, 0);
 
+    /* FIXME: accessing private data only for warning purposes...
     spice_warn_if_fail(ring_is_empty(&display->streams));
-    spice_warn_if_fail(!display->surfaces[surface_id].context.canvas);
+    */
+    spice_warn_if_fail(!display_channel_surface_has_canvas(display, surface_id));
 
     cursor_channel_reset(worker->cursor_channel);
 }
@@ -762,7 +611,7 @@ static void handle_dev_stop(void *opaque, void *payload)
 {
     RedWorker *worker = opaque;
 
-    spice_info("stop");
+    spice_debug("stop");
     spice_assert(worker->running);
 
     worker->running = FALSE;
@@ -793,10 +642,12 @@ static void handle_dev_start(void *opaque, void *payload)
 
     spice_assert(!worker->running);
     if (worker->cursor_channel) {
-        COMMON_GRAPHICS_CHANNEL(worker->cursor_channel)->during_target_migrate = FALSE;
+        CommonGraphicsChannel *common = COMMON_GRAPHICS_CHANNEL(worker->cursor_channel);
+        common_graphics_channel_set_during_target_migrate(common, FALSE);
     }
     if (worker->display_channel) {
-        worker->display_channel->common.during_target_migrate = FALSE;
+        CommonGraphicsChannel *common = COMMON_GRAPHICS_CHANNEL(worker->display_channel);
+        common_graphics_channel_set_during_target_migrate(common, FALSE);
         display_channel_wait_for_migrate_data(worker->display_channel);
     }
     worker->running = TRUE;
@@ -807,7 +658,7 @@ static void handle_dev_wakeup(void *opaque, void *payload)
 {
     RedWorker *worker = opaque;
 
-    stat_inc_counter(reds, worker->wakeup_counter, 1);
+    stat_inc_counter(worker->wakeup_counter, 1);
     red_qxl_clear_pending(worker->qxl->st, RED_DISPATCHER_PENDING_WAKEUP);
 }
 
@@ -816,16 +667,12 @@ static void handle_dev_oom(void *opaque, void *payload)
     RedWorker *worker = opaque;
     DisplayChannel *display = worker->display_channel;
 
-    RedChannel *display_red_channel = &display->common.base;
+    RedChannel *display_red_channel = RED_CHANNEL(display);
     int ring_is_empty;
 
     spice_return_if_fail(worker->running);
     // streams? but without streams also leak
-    spice_debug("OOM1 #draw=%u, #glz_draw=%u current %u pipes %u",
-                display->drawable_count,
-                display->encoder_shared_data.glz_drawable_count,
-                display->current_size,
-                red_channel_sum_pipes_size(display_red_channel));
+    display_channel_debug_oom(display, "OOM1");
     while (red_process_display(worker, &ring_is_empty)) {
         red_channel_push(display_red_channel);
     }
@@ -833,11 +680,7 @@ static void handle_dev_oom(void *opaque, void *payload)
         display_channel_free_some(worker->display_channel);
         red_qxl_flush_resources(worker->qxl);
     }
-    spice_debug("OOM2 #draw=%u, #glz_draw=%u current %u pipes %u",
-                display->drawable_count,
-                display->encoder_shared_data.glz_drawable_count,
-                display->current_size,
-                red_channel_sum_pipes_size(display_red_channel));
+    display_channel_debug_oom(display, "OOM2");
     red_qxl_clear_pending(worker->qxl->st, RED_DISPATCHER_PENDING_OOM);
 }
 
@@ -851,9 +694,8 @@ static void handle_dev_reset_cursor(void *opaque, void *payload)
 static void handle_dev_reset_image_cache(void *opaque, void *payload)
 {
     RedWorker *worker = opaque;
-    DisplayChannel *display = worker->display_channel;
 
-    image_cache_reset(&display->image_cache);
+    display_channel_reset_image_cache(worker->display_channel);
 }
 
 static void handle_dev_destroy_surface_wait_async(void *opaque, void *payload)
@@ -888,11 +730,10 @@ static void handle_dev_display_connect(void *opaque, void *payload)
     DisplayChannel *display = worker->display_channel;
     DisplayChannelClient *dcc;
 
-    spice_info("connect new client");
+    spice_debug("connect new client");
     spice_return_if_fail(display);
 
-    dcc = dcc_new(display, msg->client, msg->stream, msg->migration,
-                  msg->common_caps, msg->num_common_caps, msg->caps, msg->num_caps,
+    dcc = dcc_new(display, msg->client, msg->stream, msg->migration, &msg->caps,
                   worker->image_compression, worker->jpeg_state, worker->zlib_glz_state);
     if (!dcc) {
         return;
@@ -901,8 +742,7 @@ static void handle_dev_display_connect(void *opaque, void *payload)
     guest_set_client_capabilities(worker);
     dcc_start(dcc);
 
-    free(msg->caps);
-    free(msg->common_caps);
+    red_channel_capabilities_reset(&msg->caps);
 }
 
 static void handle_dev_display_disconnect(void *opaque, void *payload)
@@ -911,7 +751,7 @@ static void handle_dev_display_disconnect(void *opaque, void *payload)
     RedChannelClient *rcc = msg->rcc;
     RedWorker *worker = opaque;
 
-    spice_info("disconnect display client");
+    spice_debug("disconnect display client");
     spice_assert(rcc);
 
     guest_set_client_capabilities(worker);
@@ -925,7 +765,7 @@ static void handle_dev_display_migrate(void *opaque, void *payload)
     RedWorker *worker = opaque;
 
     RedChannelClient *rcc = msg->rcc;
-    spice_info("migrate display client");
+    spice_debug("migrate display client");
     spice_assert(rcc);
     red_migrate_display(worker->display_channel, rcc);
 }
@@ -973,9 +813,9 @@ static void handle_dev_monitors_config_async(void *opaque, void *payload)
         /* TODO: raise guest bug (requires added QXL interface) */
         return;
     }
-    display_update_monitors_config(worker->display_channel, dev_monitors_config,
-                                   MIN(count, msg->max_monitors),
-                                   MIN(max_allowed, msg->max_monitors));
+    display_channel_update_monitors_config(worker->display_channel, dev_monitors_config,
+                                           MIN(count, msg->max_monitors),
+                                           MIN(max_allowed, msg->max_monitors));
     red_worker_push_monitors_config(worker);
 }
 
@@ -985,13 +825,11 @@ static void handle_dev_cursor_connect(void *opaque, void *payload)
     RedWorkerMessageCursorConnect *msg = payload;
     RedWorker *worker = opaque;
 
-    spice_info("cursor connect");
+    spice_debug("cursor connect");
     cursor_channel_connect(worker->cursor_channel,
                            msg->client, msg->stream, msg->migration,
-                           msg->common_caps, msg->num_common_caps,
-                           msg->caps, msg->num_caps);
-    free(msg->caps);
-    free(msg->common_caps);
+                           &msg->caps);
+    red_channel_capabilities_reset(&msg->caps);
 }
 
 static void handle_dev_cursor_disconnect(void *opaque, void *payload)
@@ -999,7 +837,7 @@ static void handle_dev_cursor_disconnect(void *opaque, void *payload)
     RedWorkerMessageCursorDisconnect *msg = payload;
     RedChannelClient *rcc = msg->rcc;
 
-    spice_info("disconnect cursor client");
+    spice_debug("disconnect cursor client");
     spice_return_if_fail(rcc);
     red_channel_client_disconnect(rcc);
 }
@@ -1009,7 +847,7 @@ static void handle_dev_cursor_migrate(void *opaque, void *payload)
     RedWorkerMessageCursorMigrate *msg = payload;
     RedChannelClient *rcc = msg->rcc;
 
-    spice_info("migrate cursor client");
+    spice_debug("migrate cursor client");
     cursor_channel_client_migrate(rcc);
 }
 
@@ -1017,35 +855,37 @@ static void handle_dev_set_compression(void *opaque, void *payload)
 {
     RedWorkerMessageSetCompression *msg = payload;
     RedWorker *worker = opaque;
+    SpiceImageCompression image_compression = msg->image_compression;
 
-    worker->image_compression = msg->image_compression;
-    switch (worker->image_compression) {
+    switch (image_compression) {
     case SPICE_IMAGE_COMPRESSION_AUTO_LZ:
-        spice_info("ic auto_lz");
+        spice_debug("ic auto_lz");
         break;
     case SPICE_IMAGE_COMPRESSION_AUTO_GLZ:
-        spice_info("ic auto_glz");
+        spice_debug("ic auto_glz");
         break;
     case SPICE_IMAGE_COMPRESSION_QUIC:
-        spice_info("ic quic");
+        spice_debug("ic quic");
         break;
 #ifdef USE_LZ4
     case SPICE_IMAGE_COMPRESSION_LZ4:
-        spice_info("ic lz4");
+        spice_debug("ic lz4");
         break;
 #endif
     case SPICE_IMAGE_COMPRESSION_LZ:
-        spice_info("ic lz");
+        spice_debug("ic lz");
         break;
     case SPICE_IMAGE_COMPRESSION_GLZ:
-        spice_info("ic glz");
+        spice_debug("ic glz");
         break;
     case SPICE_IMAGE_COMPRESSION_OFF:
-        spice_info("ic off");
+        spice_debug("ic off");
         break;
     default:
         spice_warning("ic invalid");
+        image_compression = worker->image_compression;
     }
+    worker->image_compression = image_compression;
 
     display_channel_compress_stats_print(worker->display_channel);
     display_channel_compress_stats_reset(worker->display_channel);
@@ -1073,7 +913,7 @@ static void handle_dev_set_mouse_mode(void *opaque, void *payload)
     RedWorkerMessageSetMouseMode *msg = payload;
     RedWorker *worker = opaque;
 
-    spice_info("mouse mode %u", msg->mode);
+    spice_debug("mouse mode %u", msg->mode);
     cursor_channel_set_mouse_mode(worker->cursor_channel, msg->mode);
 }
 
@@ -1129,33 +969,26 @@ static
 void handle_dev_gl_draw_async(void *opaque, void *payload)
 {
     RedWorker *worker = opaque;
-    SpiceMsgDisplayGlDraw *draw = payload;
+    RedWorkerMessageGlDraw *draw = payload;
 
-    display_channel_gl_draw(worker->display_channel, draw);
+    display_channel_gl_draw(worker->display_channel, &draw->draw);
 }
 
-static int loadvm_command(RedWorker *worker, QXLCommandExt *ext)
+static void handle_dev_close(void *opaque, void *payload)
 {
-    RedCursorCmd *cursor_cmd;
-    RedSurfaceCmd *surface_cmd;
+    RedWorker *worker = opaque;
+    g_main_loop_quit(worker->loop);
+}
 
+static bool loadvm_command(RedWorker *worker, QXLCommandExt *ext)
+{
     switch (ext->cmd.type) {
     case QXL_CMD_CURSOR:
-        cursor_cmd = spice_new0(RedCursorCmd, 1);
-        if (red_get_cursor_cmd(&worker->mem_slots, ext->group_id, cursor_cmd, ext->cmd.data)) {
-            free(cursor_cmd);
-            return FALSE;
-        }
-        cursor_channel_process_cmd(worker->cursor_channel, cursor_cmd);
-        break;
+        return red_process_cursor_cmd(worker, ext);
+
     case QXL_CMD_SURFACE:
-        surface_cmd = spice_new0(RedSurfaceCmd, 1);
-        if (red_get_surface_cmd(&worker->mem_slots, ext->group_id, surface_cmd, ext->cmd.data)) {
-            free(surface_cmd);
-            return FALSE;
-        }
-        display_channel_process_surface_cmd(worker->display_channel, surface_cmd, TRUE);
-        break;
+        return red_process_surface_cmd(worker, ext, TRUE);
+
     default:
         spice_warning("unhandled loadvm command type (%d)", ext->cmd.type);
     }
@@ -1171,7 +1004,7 @@ static void handle_dev_loadvm_commands(void *opaque, void *payload)
     uint32_t count = msg->count;
     QXLCommandExt *ext = msg->ext;
 
-    spice_info("loadvm_commands");
+    spice_debug("loadvm_commands");
     for (i = 0 ; i < count ; ++i) {
         if (!loadvm_command(worker, &ext[i])) {
             /* XXX allow failure in loadvm? */
@@ -1187,7 +1020,7 @@ static void worker_handle_dispatcher_async_done(void *opaque,
     RedWorker *worker = opaque;
     RedWorkerMessageAsync *msg_async = payload;
 
-    spice_debug(NULL);
+    spice_debug("trace");
     red_qxl_async_complete(worker->qxl, msg_async->cmd);
 }
 
@@ -1195,7 +1028,7 @@ static void worker_dispatcher_record(void *opaque, uint32_t message_type, void *
 {
     RedWorker *worker = opaque;
 
-    red_record_event(worker->record, 1, message_type, spice_get_monotonic_time_ns());
+    red_record_event(worker->record, 1, message_type);
 }
 
 static void register_callbacks(Dispatcher *dispatcher)
@@ -1383,7 +1216,12 @@ static void register_callbacks(Dispatcher *dispatcher)
     dispatcher_register_handler(dispatcher,
                                 RED_WORKER_MESSAGE_GL_DRAW_ASYNC,
                                 handle_dev_gl_draw_async,
-                                sizeof(SpiceMsgDisplayGlDraw),
+                                sizeof(RedWorkerMessageGlDraw),
+                                DISPATCHER_NONE);
+    dispatcher_register_handler(dispatcher,
+                                RED_WORKER_MESSAGE_CLOSE_WORKER,
+                                handle_dev_close,
+                                sizeof(RedWorkerMessageClose),
                                 DISPATCHER_NONE);
 }
 
@@ -1468,7 +1306,6 @@ RedWorker* red_worker_new(QXLInstance *qxl,
     QXLDevInitInfo init_info;
     RedWorker *worker;
     Dispatcher *dispatcher;
-    const char *record_filename;
     RedsState *reds = red_qxl_get_server(qxl->st);
     RedChannel *channel;
 
@@ -1478,10 +1315,7 @@ RedWorker* red_worker_new(QXLInstance *qxl,
     worker->core = event_loop_core;
     worker->core.main_context = g_main_context_new();
 
-    record_filename = getenv("SPICE_WORKER_RECORD_FILENAME");
-    if (record_filename) {
-        worker->record = red_record_new(record_filename);
-    }
+    worker->record = reds_get_record(reds);
     dispatcher = red_qxl_get_dispatcher(qxl);
     dispatcher_set_opaque(dispatcher, worker);
 
@@ -1495,13 +1329,11 @@ RedWorker* red_worker_new(QXLInstance *qxl,
     worker->jpeg_state = reds_get_jpeg_state(reds);
     worker->zlib_glz_state = reds_get_zlib_glz_state(reds);
     worker->driver_cap_monitors_config = 0;
-#ifdef RED_STATISTICS
     char worker_str[20];
     sprintf(worker_str, "display[%d]", worker->qxl->id);
-    worker->stat = stat_add_node(reds, INVALID_STAT_REF, worker_str, TRUE);
-    worker->wakeup_counter = stat_add_counter(reds, worker->stat, "wakeups", TRUE);
-    worker->command_counter = stat_add_counter(reds, worker->stat, "commands", TRUE);
-#endif
+    stat_init_node(&worker->stat, reds, NULL, worker_str, TRUE);
+    stat_init_counter(&worker->wakeup_counter, reds, &worker->stat, "wakeups", TRUE);
+    stat_init_counter(&worker->command_counter, reds, &worker->stat, "commands", TRUE);
 
     worker->dispatch_watch =
         worker->core.watch_add(&worker->core, dispatcher_get_recv_fd(dispatcher),
@@ -1520,48 +1352,48 @@ RedWorker* red_worker_new(QXLInstance *qxl,
                       init_info.memslot_id_bits,
                       init_info.internal_groupslot_id);
 
-    spice_warn_if_fail(init_info.n_surfaces <= NUM_SURFACES);
-
     worker->event_timeout = INF_EVENT_WAIT;
 
-    worker->cursor_channel = cursor_channel_new(worker);
+    worker->cursor_channel = cursor_channel_new(reds, qxl,
+                                                &worker->core);
     channel = RED_CHANNEL(worker->cursor_channel);
+    red_channel_init_stat_node(channel, &worker->stat, "cursor_channel");
     red_channel_register_client_cbs(channel, client_cursor_cbs, dispatcher);
+    g_object_set_data(G_OBJECT(channel), "dispatcher", dispatcher);
     reds_register_channel(reds, channel);
 
     // TODO: handle seemless migration. Temp, setting migrate to FALSE
-    worker->display_channel = display_channel_new(reds, worker, FALSE,
+    worker->display_channel = display_channel_new(reds, qxl, &worker->core, FALSE,
                                                   reds_get_streaming_video(reds),
                                                   reds_get_video_codecs(reds),
                                                   init_info.n_surfaces);
-
     channel = RED_CHANNEL(worker->display_channel);
+    red_channel_init_stat_node(channel, &worker->stat, "display_channel");
     red_channel_register_client_cbs(channel, client_display_cbs, dispatcher);
-    red_channel_set_cap(channel, SPICE_DISPLAY_CAP_MONITORS_CONFIG);
-    red_channel_set_cap(channel, SPICE_DISPLAY_CAP_PREF_COMPRESSION);
-    red_channel_set_cap(channel, SPICE_DISPLAY_CAP_STREAM_REPORT);
+    g_object_set_data(G_OBJECT(channel), "dispatcher", dispatcher);
     reds_register_channel(reds, channel);
 
     return worker;
 }
 
-SPICE_GNUC_NORETURN static void *red_worker_main(void *arg)
+static void *red_worker_main(void *arg)
 {
     RedWorker *worker = arg;
 
-    spice_info("begin");
-    spice_assert(MAX_PIPE_SIZE > WIDE_CLIENT_ACK_WINDOW &&
+    spice_debug("begin");
+    SPICE_VERIFY(MAX_PIPE_SIZE > WIDE_CLIENT_ACK_WINDOW &&
            MAX_PIPE_SIZE > NARROW_CLIENT_ACK_WINDOW); //ensure wakeup by ack message
 
-    RED_CHANNEL(worker->cursor_channel)->thread_id = pthread_self();
-    RED_CHANNEL(worker->display_channel)->thread_id = pthread_self();
+    red_channel_reset_thread_id(RED_CHANNEL(worker->cursor_channel));
+    red_channel_reset_thread_id(RED_CHANNEL(worker->display_channel));
 
     GMainLoop *loop = g_main_loop_new(worker->core.main_context, FALSE);
+    worker->loop = loop;
     g_main_loop_run(loop);
     g_main_loop_unref(loop);
+    worker->loop = NULL;
 
-    /* FIXME: free worker, and join threads */
-    exit(0);
+    return NULL;
 }
 
 bool red_worker_run(RedWorker *worker)
@@ -1586,7 +1418,41 @@ bool red_worker_run(RedWorker *worker)
     return r == 0;
 }
 
-static RedsState* red_worker_get_server(RedWorker *worker)
+static void red_worker_close_channel(RedChannel *channel)
 {
-    return red_qxl_get_server(worker->qxl->st);
+    red_channel_reset_thread_id(channel);
+    red_channel_destroy(channel);
+}
+
+/*
+ * Free the worker thread. This function should be called by RedQxl
+ * after sending a RED_WORKER_MESSAGE_CLOSE_WORKER message;
+ * failing to do so will cause a deadlock.
+ */
+void red_worker_free(RedWorker *worker)
+{
+    RedsState *reds = red_qxl_get_server(worker->qxl->st);
+
+    /* prevent any possible future attempt to connect to new clients */
+    reds_unregister_channel(reds, RED_CHANNEL(worker->cursor_channel));
+    reds_unregister_channel(reds, RED_CHANNEL(worker->display_channel));
+
+    pthread_join(worker->thread, NULL);
+
+    red_worker_close_channel(RED_CHANNEL(worker->cursor_channel));
+    worker->cursor_channel = NULL;
+    red_worker_close_channel(RED_CHANNEL(worker->display_channel));
+    worker->display_channel = NULL;
+
+    if (worker->dispatch_watch) {
+        worker->core.watch_remove(&worker->core, worker->dispatch_watch);
+    }
+
+    g_main_context_unref(worker->core.main_context);
+
+    if (worker->record) {
+        red_record_unref(worker->record);
+    }
+    memslot_info_destroy(&worker->mem_slots);
+    free(worker);
 }
