@@ -15,8 +15,9 @@
    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
-#ifndef REDS_PRIVATE_H
-#define REDS_PRIVATE_H
+
+#ifndef REDS_PRIVATE_H_
+#define REDS_PRIVATE_H_
 
 #include <spice/protocol.h>
 #include <spice/stats.h>
@@ -24,6 +25,8 @@
 #include "main-dispatcher.h"
 #include "main-channel.h"
 #include "inputs-channel.h"
+#include "stat-file.h"
+#include "red-record-qxl.h"
 
 #define MIGRATE_TIMEOUT (MSEC_PER_SEC * 10)
 #define MM_TIME_DELTA 400 /*ms*/
@@ -45,43 +48,17 @@ typedef struct MonitorMode {
     uint32_t y_res;
 } MonitorMode;
 
-#ifdef RED_STATISTICS
-
-#define REDS_MAX_STAT_NODES 100
-#define REDS_STAT_SHM_SIZE (sizeof(SpiceStat) + REDS_MAX_STAT_NODES * sizeof(SpiceStatNode))
-
-typedef struct RedsStatValue {
-    uint32_t value;
-    uint32_t min;
-    uint32_t max;
-    uint32_t average;
-    uint32_t count;
-} RedsStatValue;
-
-#endif
-
 typedef struct RedsMigPendingLink {
-    RingItem ring_link; // list of links that belongs to the same client
     SpiceLinkMess *link_msg;
     RedsStream *stream;
 } RedsMigPendingLink;
 
 typedef struct RedsMigTargetClient {
     RedsState *reds;
-    RingItem link;
     RedClient *client;
-    Ring pending_links;
+    GList *pending_links;
 } RedsMigTargetClient;
 
-/* Intermediate state for on going monitors config message from a single
- * client, being passed to the guest */
-typedef struct RedsClientMonitorsConfig {
-    MainChannelClient *mcc;
-    uint8_t *buffer;
-    int buffer_size;
-    int buffer_pos;
-} RedsClientMonitorsConfig;
-
 typedef struct ChannelSecurityOptions ChannelSecurityOptions;
 
 typedef struct RedSSLParameters {
@@ -104,8 +81,7 @@ struct RedsState {
     SpiceWatch *secure_listen_watch;
     RedCharDeviceVDIPort *agent_dev;
     int pending_mouse_event;
-    Ring clients;
-    int num_clients;
+    GList *clients;
     MainChannel *main_channel;
     InputsChannel *inputs_channel;
 
@@ -122,10 +98,9 @@ struct RedsState {
                                     between the 2 servers */
     int dst_do_seamless_migrate; /* per migration. Updated after the migration handshake
                                     between the 2 servers */
-    Ring mig_target_clients;
+    GList *mig_target_clients;
 
-    int num_of_channels;
-    Ring channels;
+    GList *channels;
     int mouse_mode;
     int is_client_mouse_allowed;
     int dispatcher_allows_client_mouse;
@@ -139,23 +114,27 @@ struct RedsState {
     SSL_CTX *ctx;
 
 #ifdef RED_STATISTICS
-    char *stat_shm_name;
-    SpiceStat *stat;
-    pthread_mutex_t stat_lock;
-    RedsStatValue roundtrip_stat;
+    RedStatFile *stat_file;
 #endif
     int allow_multiple_clients;
 
-    RedsClientMonitorsConfig client_monitors_config;
+    /* Intermediate state for on going monitors config message from a single
+     * client, being passed to the guest */
+    SpiceBuffer client_monitors_config;
+
     int mm_time_enabled;
     uint32_t mm_time_latency;
 
     SpiceCharDeviceInstance *vdagent;
     SpiceMigrateInstance *migration_interface;
 
-    SpiceCoreInterfaceInternal *core;
+    SpiceCoreInterfaceInternal core;
     GList *qxl_instances;
     MainDispatcher *main_dispatcher;
+    RedRecord *record;
 };
 
-#endif
+#define FOREACH_QXL_INSTANCE(_reds, _iter, _qxl) \
+    GLIST_FOREACH(_reds->qxl_instances, _iter, QXLInstance, _qxl)
+
+#endif /* REDS_PRIVATE_H_ */
