@@ -15,8 +15,8 @@
    License along with this library; if not, see <http://www.gnu.org/licenses/>.
 */
 
-#ifndef _H_REDS
-#define _H_REDS
+#ifndef REDS_H_
+#define REDS_H_
 
 #include <stdint.h>
 #include <sys/uio.h>
@@ -47,10 +47,11 @@ void reds_enable_mm_time(RedsState *reds);
 uint32_t reds_get_mm_time(void);
 void reds_register_channel(RedsState *reds, RedChannel *channel);
 void reds_unregister_channel(RedsState *reds, RedChannel *channel);
+RedChannel *reds_find_channel(RedsState *reds, uint32_t type, uint32_t id);
 int reds_get_mouse_mode(RedsState *reds); // used by inputs_channel
 gboolean reds_config_get_agent_mouse(const RedsState *reds); // used by inputs_channel
 int reds_has_vdagent(RedsState *reds); // used by inputs channel
-int reds_config_get_playback_compression(RedsState *reds); // used by playback channel
+bool reds_config_get_playback_compression(RedsState *reds); // used by playback channel
 
 void reds_handle_agent_mouse_event(RedsState *reds, const VDAgentMouseState *mouse_state); // used by inputs_channel
 
@@ -78,11 +79,12 @@ void reds_on_main_agent_start(RedsState *reds, MainChannelClient *mcc, uint32_t
 void reds_on_main_agent_tokens(RedsState *reds, MainChannelClient *mcc, uint32_t num_tokens);
 uint8_t *reds_get_agent_data_buffer(RedsState *reds, MainChannelClient *mcc, size_t size);
 void reds_release_agent_data_buffer(RedsState *reds, uint8_t *buf);
-void reds_on_main_agent_data(RedsState *reds, MainChannelClient *mcc, void *message, size_t size);
+void reds_on_main_agent_data(RedsState *reds, MainChannelClient *mcc, const void *message,
+                             size_t size);
 void reds_on_main_migrate_connected(RedsState *reds, int seamless); //should be called when all the clients
                                                    // are connected to the target
-int reds_handle_migrate_data(RedsState *recs, MainChannelClient *mcc,
-                             SpiceMigrateDataMain *mig_data, uint32_t size);
+bool reds_handle_migrate_data(RedsState *recs, MainChannelClient *mcc,
+                              SpiceMigrateDataMain *mig_data, uint32_t size);
 void reds_on_main_mouse_mode_request(RedsState *reds, void *message, size_t size);
 /* migration dest side: returns whether it can support seamless migration
  * with the given src migration protocol version */
@@ -100,6 +102,11 @@ SpiceCoreInterfaceInternal* reds_get_core_interface(RedsState *reds);
 void reds_update_client_mouse_allowed(RedsState *reds);
 MainDispatcher* reds_get_main_dispatcher(RedsState *reds);
 
+/* Get the recording object stored in RedsState.
+ * You should free with red_record_unref.
+ */
+struct RedRecord *reds_get_record(RedsState *reds);
+
 /* fd watches/timers */
 SpiceWatch *reds_core_watch_add(RedsState *reds,
                                 int fd, int event_mask,
@@ -121,4 +128,4 @@ void reds_core_timer_cancel(RedsState *reds,
 void reds_core_timer_remove(RedsState *reds,
                             SpiceTimer *timer);
 
-#endif
+#endif /* REDS_H_ */
